/*
  GaussianND.cpp

  (C) 2005 Paul Vernaza, <vernaza@stwing.upenn.edu>
*/

#include "GaussianND.h"
 
#include <assert.h>
#include <math.h>
#include <string.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_linalg.h>

#define MAX_DIM 30

GaussianND::GaussianND(int dim_a) {
  dim = dim_a;
  
  covarCopy = (double*)malloc(sizeof(double)*dim*dim);
  invCovar = (double*)malloc(sizeof(double)*dim*dim);
  scaler = pow(M_PI * 2.0, dim / 2.0);
  
  gslPerm = gsl_permutation_alloc(dim);
}

GaussianND::GaussianND(GaussianND &src) {
  dim = src.dim;
  scaler = src.scaler;
  
  covarCopy = (double*)malloc(sizeof(double)*dim*dim);
  invCovar = (double*)malloc(sizeof(double)*dim*dim);
  
  mean = src.mean;
  covar = src.covar;
  gslPerm = gsl_permutation_alloc(dim);
}

void GaussianND::SetCovar(double *covar_a) {
  covar = covar_a;
}
 
void GaussianND::LnValue(double *vector, 
			 double *exponent_out, double *normalizer_out) {
  int signum;
 
  double deviation[MAX_DIM];
 
  for (int i = 0; i < dim; i++) {
    deviation[i] = vector[i] - mean[i];
    assert(!isnan(vector[i]));
  }
 
  memcpy(covarCopy, covar, sizeof(double)*dim*dim);
 
  gsl_matrix_view gsl_covar = gsl_matrix_view_array(covarCopy, dim, dim);
  gsl_matrix_view gsl_inv_covar = gsl_matrix_view_array(invCovar, dim, dim);
 
  gsl_linalg_LU_decomp(&gsl_covar.matrix, gslPerm, &signum);
 
  gsl_linalg_LU_invert(&gsl_covar.matrix, gslPerm, &gsl_inv_covar.matrix);
 
  double root_covar_det = 
    sqrt(gsl_linalg_LU_det(&gsl_covar.matrix, signum));
 
  //  printf("root covar det = %.32lf\n", root_covar_det);
 
  double product_temp[MAX_DIM];
 
  cblas_dsymv(CblasRowMajor, CblasUpper, dim, 
 	      1.0, invCovar, dim, 
 	      deviation, 1, 
 	      0.0, product_temp, 1);
 
  *exponent_out = -0.5 * cblas_ddot(dim, deviation, 1, product_temp, 1);
 
  //  printf("parameter = %f\n", parameter);
 
  *normalizer_out = 1 / (scaler * root_covar_det);
 
  //  printf("normalizer = %.32lf\n", normalizer);
}
 
double GaussianND::LnValue(double *vector) {
  double exponent, normalizer;
  LnValue(vector, &exponent, &normalizer);
 
  double log_norm = log(normalizer);
 
  if (isnan(log(normalizer))) 
    log_norm = 0.0;
 
  //  assert(!isnan(log(normalizer)));
  assert(!isnan(exponent));
  return log_norm + exponent;
}
 
double GaussianND::Value(double *vector) {
  double exponent, normalizer;
  LnValue(vector, &exponent, &normalizer);
  return normalizer * exp(exponent);
}
 
/*
  int main(int argc, char *argv[]) {
  double covar[3][3] = { 1.0, 0.0, 0.0,
  0.0, 4.0, 0.0,
  0.0, 0.0, 9.0 };
 
  //  memset(covar, 0, sizeof(double)*9);
 
  double mean[3] = { 4.0, 7.0, 1.0 };
 
  GaussianND gaussian(3);
 
  gaussian.SetMean(mean);
  gaussian.SetCovar(&covar[0][0]);
 
  double vector[3] = { 1.0, 2.0, 3.0 };
 
  double value = gaussian.Value(vector);
 
  printf("Gaussian = %.4e\n", value);
  }
*/
 
