/* 
   KalmanUpdater.cpp

   (C) 2005 Paul Vernaza, <vernaza@stwing.upenn.edu>
*/

#include "KalmanUpdater.h"

/*
extern "C" {
#include <cblas.h>
}
*/

#include <gsl/gsl_linalg.h>
#include <gsl/gsl_cblas.h>
  
#define MAX_STATE_DIM 20
#define MAX_MEASUREMENT_DIM 20

KalmanUpdater::KalmanUpdater(int state_dim, int measurement_dim,
			     double *process_matrix, 
			     double *process_covar, 
			     double *measurement_matrix,
			     double *measurement_covar) {
  stateDim = state_dim;
  measurementDim = measurement_dim;

  gslPerm = gsl_permutation_alloc(measurement_dim);

  processMatrix = process_matrix;
  processCovar = process_covar;
  measurementMatrix = measurement_matrix;
  measurementCovar = measurement_covar;

  measurementLikelihoodMean = (double*)malloc(sizeof(double)*measurement_dim);
  measurementLikelihoodCovar = 
    (double*)malloc(sizeof(double)*measurement_dim*measurement_dim);
  kalmanGain = (double*)malloc(sizeof(double)*state_dim*measurement_dim);
}

void KalmanUpdater::ProcessUpdateCovar(double *covar, 
				       double *updated_prior_covar) {
  double prior_covar_temp[MAX_STATE_DIM*MAX_STATE_DIM];

  /*
  printf("\nprocess\n");
  for (int i = 0; i < stateDim; i++) {
    //    printf("[ ");
    for (int j = 0; j < stateDim; j++) {
      printf("%8g ", processMatrix[i*stateDim + j]);
    }
    //    printf("] \n");
    printf(";\n");
  }

  printf("prior\n");
  for (int i = 0; i < stateDim; i++) {
    //    printf("[ ");
    for (int j = 0; j < stateDim; j++) {
      printf("%8g ", covar[i*stateDim + j]);
    }
    //    printf("] \n");
    printf("\n;");
  }
  */

  // prior_covar_temp = processMatrix * covar
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
	      stateDim, stateDim, stateDim, 
	      1.0, processMatrix, stateDim, 
	      covar, stateDim, 
	      0.0, prior_covar_temp, stateDim);

  memcpy(updated_prior_covar, processCovar, 
	 sizeof(double) * stateDim * stateDim);

  // updated_prior_covar = prior_covar_temp*processMatrix' + process_covar
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasTrans, 
	      stateDim, stateDim, stateDim,
	      1.0, prior_covar_temp, stateDim, 
	      processMatrix, stateDim, 
	      1.0, updated_prior_covar, stateDim);

  /*
  printf("post\n");
  for (int i = 0; i < stateDim; i++) {
    //    printf("[ ");
    printf(" ");
    for (int j = 0; j < stateDim; j++) {
      printf("%8g ", updated_prior_covar[i*stateDim + j]);
    }
    //    printf("] \n");
    printf("\n;");
  }

  printf("kyoo\n");
  for (int i = 0; i < stateDim; i++) {
    //    printf("[ ");
    printf(" ");
    for (int j = 0; j < stateDim; j++) {
      printf("%8g ", processCovar[i*stateDim + j]);
    }
    //    printf("] \n");
    printf("\n;");
  }
  */

}

void KalmanUpdater::UpdateCovar(double *covar, double *covar_out) {
  double updated_prior_covar[MAX_STATE_DIM*MAX_STATE_DIM];
  ProcessUpdateCovar(covar, updated_prior_covar);
  CalcKalmanGain(updated_prior_covar, kalmanGain);
  CalcPosteriorCovar(kalmanGain, updated_prior_covar, covar_out);
}

void KalmanUpdater::CalcKalmanGain(double *updated_prior_covar, 
				   double *kalman_gain_out) {
  double kalman_gain_temp_PBt[MAX_STATE_DIM * MAX_MEASUREMENT_DIM];
  double kalman_gain_temp_BP[MAX_STATE_DIM * MAX_MEASUREMENT_DIM];

  // kalman_gain_temp_PBt = updated_prior_covar * measurementMatrix'
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasTrans, 
	      stateDim, measurementDim, stateDim, 
	      1.0, updated_prior_covar, stateDim,
	      measurementMatrix, stateDim,
	      0.0, kalman_gain_temp_PBt, measurementDim);

  //  pretty_print_matrix("PB'", kalman_gain_temp_PBt, stateDim, measurementDim);

  // kalman_gain_temp_BP = measurementMatrix * updated_prior_covar
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
	      measurementDim, stateDim, stateDim,
	      1.0, measurementMatrix, stateDim,
	      updated_prior_covar, stateDim,
	      0.0, kalman_gain_temp_BP, stateDim);
  
  memcpy(measurementLikelihoodCovar, measurementCovar, 
	 sizeof(double) * measurementDim * measurementDim);

  // measurementLikelihoodCovar = measurementCovar +
  //    kalman_gain_temp_BP * measurementMatrix'
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasTrans,
	      measurementDim, measurementDim, stateDim,
	      1.0, kalman_gain_temp_BP, stateDim,
	      measurementMatrix, stateDim, 
	      1.0, measurementLikelihoodCovar, measurementDim);
  
  double measLikelihoodCopy[MAX_MEASUREMENT_DIM][MAX_MEASUREMENT_DIM];
  memcpy(&measLikelihoodCopy[0][0], measurementLikelihoodCovar, 
	 sizeof(double)*measurementDim*measurementDim);

  gsl_matrix_view gsl_m_likelihood_copy = 
    gsl_matrix_view_array(&measLikelihoodCopy[0][0], 
			  measurementDim, measurementDim);
  double m_likelihood_inv[MAX_MEASUREMENT_DIM * MAX_MEASUREMENT_DIM];
  gsl_matrix_view gsl_m_likelihood_inv = 
    gsl_matrix_view_array(m_likelihood_inv, measurementDim, measurementDim);
  
  int signum;

  gsl_linalg_LU_decomp(&gsl_m_likelihood_copy.matrix, gslPerm, &signum);

  gsl_linalg_LU_invert(&gsl_m_likelihood_copy.matrix, gslPerm, &gsl_m_likelihood_inv.matrix);
  
  // kalman_gain_out = kalman_gain_temp_PBt * m_likelihood_inv
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
	      stateDim, measurementDim, measurementDim,
	      1.0, kalman_gain_temp_PBt, measurementDim,
	      m_likelihood_inv, measurementDim,
	      0.0, kalman_gain_out, measurementDim);

  //  pretty_print_matrix("kalman gain", kalman_gain_out, stateDim, measurementDim);

}

void KalmanUpdater::CalcPosteriorCovar(double *kalman_gain, 
				       double *updated_prior_covar,
				       double *covar_out) {
  //  double KB[MAX_STATEDIM * MAX_STATE_DIM];
  double I_minus_KB[MAX_STATE_DIM * MAX_STATE_DIM];
  memset(I_minus_KB, 0, sizeof(double)*stateDim*stateDim);

  for (int i = 0; i < stateDim; i++) 
    I_minus_KB[i + i * stateDim] = 1.0;

  // I_minus_KB = I - kalman_gain * measurementMatrix
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
	      stateDim, stateDim, measurementDim, 
	      -1.0, kalman_gain, measurementDim, 
	      measurementMatrix, stateDim,
	      1.0, I_minus_KB, stateDim);

  double covar_temp[MAX_STATE_DIM * MAX_STATE_DIM];

  // covar_temp = I_minus_KB * updated_prior_covar
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
	      stateDim, stateDim, stateDim, 
	      1.0, I_minus_KB, stateDim,
	      updated_prior_covar, stateDim,
	      0.0, covar_temp, stateDim);

  //  pretty_print_matrix("eye minus kb", I_minus_KB, state_dim, state_dim);

  // covar_out = covar_temp * I_minus_KB'
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasTrans,
	      stateDim, stateDim, stateDim,
	      1.0, covar_temp, stateDim,
	      I_minus_KB, stateDim,
	      0.0, covar_out, stateDim);

  double KR[MAX_STATE_DIM * MAX_MEASUREMENT_DIM];
  // KR = kalman_gain * measurementCovar
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
	      stateDim, measurementDim, measurementDim,
	      1.0, kalman_gain, measurementDim,
	      measurementCovar, measurementDim, 
	      0.0, KR, measurementDim);

  // covar_out = KR * kalman_gain' + covar_out
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasTrans,
	      stateDim, stateDim, measurementDim,
	      1.0, KR, measurementDim,
	      kalman_gain, measurementDim,
	      1.0, covar_out, stateDim);
}

void KalmanUpdater::Update(double *measurement, 
			   double *mean, double *covar,
			   double *mean_out, double *covar_out) {
  double covar_buf[MAX_STATE_DIM * MAX_STATE_DIM];
  double mean_buf[MAX_STATE_DIM];
  bool use_buffer = false;

  if (mean_out == NULL) {
    mean_out = mean_buf;
    covar_out = covar_buf;
    use_buffer = true;
  }

  UpdateCovar(covar, covar_out);
  UpdateMean(measurement, mean, mean_out);

  if (use_buffer) {
    memcpy(mean, mean_out, sizeof(double)*stateDim);
    memcpy(covar, covar_out, sizeof(double)*stateDim*stateDim);
  }
}

void KalmanUpdater::ProcessUpdateMean(double *mean, double *mean_out) {
  // mean_out = processMatrix * mean
  cblas_dgemv(CblasRowMajor, CblasNoTrans, 
	      stateDim, stateDim,
	      1.0, processMatrix, stateDim, 
	      mean, 1, 
	      0.0, mean_out, 1);
}

void KalmanUpdater::UpdateMean(double *measurement, 
			       double *mean, double *mean_out) 
{
  // updated_prior = processMatrix * mean
  double updated_prior[MAX_STATE_DIM];

  ProcessUpdateMean(mean, updated_prior);

  // measurementLikelihoodMean = measurementMatrix * updated_prior
  cblas_dgemv(CblasRowMajor, CblasNoTrans, 
	      measurementDim, stateDim,
	      1.0, measurementMatrix, stateDim,
	      updated_prior, 1,
	      0.0, measurementLikelihoodMean, 1);
  
  // innovation = measurement - measurementLikelihoodMean
  double innovation[MAX_MEASUREMENT_DIM];
  memcpy(innovation, measurement, sizeof(double)*measurementDim);
  cblas_daxpy(measurementDim, 
	      -1.0, measurementLikelihoodMean, 1,
	      innovation, 1);

  // updated_prior = updated_prior + kalmanGain * innovation
  cblas_dgemv(CblasRowMajor, CblasNoTrans,
	      stateDim, measurementDim,
	      1.0, kalmanGain, measurementDim,
	      innovation, 1,
	      1.0, updated_prior, 1);

  memcpy(mean_out, updated_prior, sizeof(double)*stateDim);
}

/*
#include <unistd.h>
int main(int argc, char *argv[]) {
  int stateDim = 9;
  int measurementDim = 3;

  KalmanUpdater kf(stateDim, measurementDim);

  double delta_t = 0.01;
  double processMatrix[stateDim][stateDim];

  memset(processMatrix, 0, sizeof(double)*stateDim*stateDim);

  for (int i = 0; i < stateDim; i++) 
    processMatrix[i][i] = 1.0;

  // position
  for (int i = 0; i < 3; i++) 
    processMatrix[i][i+3] = delta_t;
  for (int i = 0; i < 3; i++) 
    processMatrix[i][i+6] = 0.5 * delta_t * delta_t;
  
  // velocity 
  for (int i = 0; i < 3; i++) 
    processMatrix[i+3][i+6] = delta_t;

  double measurementMatrix[measurementDim][stateDim];
  memset(measurementMatrix, 0, sizeof(double)*measurementDim*stateDim);

  for (int i = 0; i < measurementDim; i++) 
    measurementMatrix[i][i] = 1.0;

  double process_covar[stateDim][stateDim];
  memset(process_covar, 0, sizeof(double)*stateDim*stateDim);
  for (int i = 0; i < stateDim; i++) 
    process_covar[i][i] = 0.01;

  double measurementCovar[measurementDim][measurementDim];
  memset(measurementCovar, 0, sizeof(double)*measurementDim*measurementDim);
  for (int i = 0; i < measurementDim; i++) 
    measurementCovar[i][i] = 0.01;

  double est_mean[stateDim];
  double est_covar[stateDim][stateDim];

  memset(est_covar, 0, sizeof(double)*stateDim*stateDim);

  for (int i = 0; i < stateDim; i++) {
    est_mean[i] = 1.0;
    est_covar[i][i] = 100;
  }

  kf.SetProcessMatrix(&processMatrix[0][0]);
  kf.SetMeasurementMatrix(&measurementMatrix[0][0]);
  kf.SetProcessCovar(&process_covar[0][0]);
  kf.SetMeasurementCovar(&measurementCovar[0][0]);

  pretty_print_matrix("process matrix", &processMatrix[0][0], stateDim, stateDim);
  pretty_print_matrix("measurement matrix", &measurementMatrix[0][0], measurementDim, stateDim);
  pretty_print_matrix("process covariance", &process_covar[0][0], stateDim, stateDim);
  pretty_print_matrix("measurement covariance", &measurementCovar[0][0], measurementDim, measurementDim);

  double measurement[measurementDim];
  for (int i = 0; i < measurementDim; i++)
    measurement[i] = 5.0;

  int counter = 0;
  while (1) {
    double new_mean[stateDim], new_covar[stateDim][stateDim];

    //    pretty_print_matrix("covariance", &est_covar[0][0], 9, 9);

    printf("mean = %lf %lf %lf, %lf %lf %lf, %lf %lf %lf\n",
	   //    printf("mean = %lf %lf %lf, %lf %lf %lf\n",
	   est_mean[0], est_mean[1], est_mean[2],
	   est_mean[3], est_mean[4], est_mean[5]
	   , est_mean[6], est_mean[7], est_mean[8]
	   );

    kf.Update(measurement, est_mean, &est_covar[0][0], 
	      new_mean, &new_covar[0][0]);

    double *m_likelihood_mean, *m_likelihood_covar;
    kf.GetMeasurementLikelihoodGaussian(&m_likelihood_mean,
					&m_likelihood_covar);

    pretty_print_vector("measurement likelihood mean", m_likelihood_mean, measurementDim);
    pretty_print_matrix("measurement likelihood covar", m_likelihood_covar, measurementDim, measurementDim);

    memcpy(est_mean, new_mean, sizeof(double)*stateDim);
    memcpy(est_covar, new_covar, sizeof(double)*stateDim*stateDim);
    //    if (counter++ % 100 == 0) sleep(1);
    sleep(1);
  }
}
*/

