/* 
   KalmanUpdater.h

   (C) 2005 Paul Vernaza, <vernaza@stwing.upenn.edu>
*/

#ifndef __KALMAN_UPDATER_H
#define __KALMAN_UPDATER_H

#include <stdio.h>
#include <string.h>
#include <gsl/gsl_linalg.h>

class KalmanUpdater {
 public:
  KalmanUpdater(int state_dim, int measurement_dim,
		double *process_matrix, 
		double *process_covar, 
		double *measurement_matrix,
		double *measurement_covar);

  ~KalmanUpdater() { 
    gsl_permutation_free(gslPerm);
    free(measurementLikelihoodMean);
    free(measurementLikelihoodCovar);
    free(kalmanGain);
  }
  
  /* Do a complete Kalman update on the given mean and covariance
     given the measurement.  If mean_out and covar_out are provided,
     output stored in those buffers.  Otherwise, output is stored 
     in input buffers.
   */
  void Update(double *measurement, 
	      double *mean, double *covar,
	      double *mean_out = NULL, double *covar_out = NULL);

  /* Apply a process update to the mean */
  void ProcessUpdateMean(double *mean, double *mean_out);

  /* Apply a process update to the covariance */
  void ProcessUpdateCovar(double *covar, double *updated_prior_covar);

  /* Do a process and measurement update on the mean */
  void UpdateMean(double *measurement, double *mean, double *mean_out);

  /* Do a process and measurement update on the covariance  */
  void UpdateCovar(double *covar, double *covar_out);

  /* Return parameters of the Gaussian characterizing the likelihood 
     of the measurement in the last complete update
  */
  double *GetMeasurementLikelihoodMean() { return measurementLikelihoodMean; }
  double *GetMeasurementLikelihoodCovar() { return measurementLikelihoodCovar; }

 private:
  int stateDim, measurementDim;
  double *measurementLikelihoodMean, *measurementLikelihoodCovar;
  double *kalmanGain;
  //  int stateDim, measurementDim;
  gsl_permutation *gslPerm;
  double *processMatrix, *measurementMatrix,
    *processCovar, *measurementCovar;
  
  void CalcKalmanGain(double *updated_prior_covar, double *kalman_gain_out);
  void CalcPosteriorCovar(double *kalman_gain, double *updated_prior_covar,
			  double *covar_out);
};

#endif
