/* 
   PoseRBFilter.cpp

   (C) 2005 Paul Vernaza, <vernaza@stwing.upenn.edu>
*/

#include <algorithm>
#include <assert.h>
#include <fcntl.h>
#include <fenv.h>
#include <functional>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "DebugMatrices.h"
#include "GaussianND.h"
#include "PoseRBFilter.h"
#include "ModelMatrices.h"
#include "Quaternion.h"
#include "RBParticle.h"
#include "UTM/LatLong-UTMconversion.h"

using namespace std;

struct particle_comparer_ascending {
  bool operator()(RBParticle *a, RBParticle *b)  { 
    return (a->GetWeight() < b->GetWeight());
  }
};

struct particle_comparer_descending {
  bool operator()(RBParticle *a, RBParticle *b)  { 
    return (a->GetWeight() > b->GetWeight());
  }
};

PoseRBFilter::PoseRBFilter(int particleCount_, bool randomize_rotations) : 
  particleCount(particleCount_),
  accelMeasurePDF(3),
  velMeasurePDF(3),
  gpsVelMeasurePDF(3),
  posMeasurePDF(3),
  gravMeasurePDF(3),
  magfieldMeasurePDF(3)
{
  srand48(time(NULL));

  // allocate two buffers of pointers for double-buffering
  particleBuffer1 = (RBParticle**)malloc(sizeof(RBParticle*)*particleCount);
  particleBuffer2 = (RBParticle**)malloc(sizeof(RBParticle*)*particleCount);
  particles = particleBuffer1;
  newParticles = particleBuffer2;

  analyticalStateDim = processMatrix.Rows();

  // we use these to evaluate Kalman updates for different kinds 
  // of measurements 
  kalmanPosUpdater =
    new KalmanUpdater(analyticalStateDim, posMeasurementMatrix.Rows(),
		      processMatrix.GetMatrix(), processCovar.GetMatrix(),
		      posMeasurementMatrix.GetMatrix(), posMeasurementCovar.GetMatrix());

  kalmanOdUpdater =
    new KalmanUpdater(processMatrix.Rows(), velMeasurementMatrix.Rows(),
		      processMatrix.GetMatrix(), processCovar.GetMatrix(),
		      velMeasurementMatrix.GetMatrix(), odMeasurementCovar.GetMatrix());

  kalmanVisOdUpdater = 
    new KalmanUpdater(processMatrix.Rows(), velMeasurementMatrix.Rows(),
		      processMatrix.GetMatrix(), processCovar.GetMatrix(),
		      velMeasurementMatrix.GetMatrix(), visOdMeasurementCovar.GetMatrix());

  kalmanGPSVelUpdater =
    new KalmanUpdater(processMatrix.Rows(), gpsVelMeasurementMatrix.Rows(),
		      processMatrix.GetMatrix(), processCovar.GetMatrix(),
		      gpsVelMeasurementMatrix.GetMatrix(), gpsVelMeasurementCovar.GetMatrix());
 
  kalmanAccelUpdater =
    new KalmanUpdater(processMatrix.Rows(), accelMeasurementMatrix.Rows(),
		      processMatrix.GetMatrix(), processCovar.GetMatrix(),
		      accelMeasurementMatrix.GetMatrix(), accelMeasurementCovar.GetMatrix());

  InitialPoseMean initialPoseMean;
  InitialPoseCovar initialPoseCovar;

  // construct particles
  for (int i = 0; i < particleCount; i++) {
    particles[i] = 
      new RBParticle(analyticalStateDim, initialPoseMean.GetMatrix(), initialPoseCovar.GetMatrix());
    if (randomize_rotations)
      particles[i]->Randomize();
  }

  // get gravity and magnetic field reference vectors
  GravityRef grefmat;
  MagfieldRef mrefmat;
  memcpy(gravityRef, grefmat.GetMatrix(), sizeof(double)*3);
  memcpy(magfieldRef, mrefmat.GetMatrix(), sizeof(double)*3);

  MagfieldBias mbiasmat;
  memcpy(magfieldBias, mbiasmat.GetMatrix(), sizeof(double)*3);

  GPSPosOffset gposoffset;
  memcpy(gpsPosOffset, gposoffset.GetMatrix(), sizeof(double)*3);

  // initialize Gaussians characterizing IMU measurements on orientation 
  gravMeasurePDF.SetMean(gravityRef);
  gravMeasurePDF.SetCovar(gravMeasurementCovar.GetMatrix());

  magfieldMeasurePDF.SetMean(magfieldRef);
  magfieldMeasurePDF.SetCovar(magfieldMeasurementCovar.GetMatrix());

  estimateParticle = new RBParticle(analyticalStateDim);

  lastPosUpdateTstamp = 0.0;
  lastRotUpdateTstamp = 0.0;
  lastResampleTime = 0.0;
}

double PoseRBFilter::UpdatePosTstamp(double tstamp) {
  double delta_time, dt_squared;

  if (lastPosUpdateTstamp == 0.0) {
    delta_time = dt_squared = 0.0;
  } else {
    //    printf("TST %lf LAST %lf\n", tstamp, lastPosUpdateTstamp);
    delta_time = tstamp - lastPosUpdateTstamp;
    dt_squared = delta_time * delta_time;
  }

  lastPosUpdateTstamp = tstamp;

  // update delta_t parameter in Kalman matrices
  processMatrix.ReplaceDT(delta_time);
  //  processMatrix.ReplaceDT2(dt_squared);
  processCovar.ReplaceDT(delta_time);

  return delta_time;
}

double PoseRBFilter::UpdateRotTstamp(double tstamp) {
  double delta_time;

  if (lastRotUpdateTstamp == 0.0) {
    delta_time = 0.0;
  } else {
    delta_time = tstamp - lastRotUpdateTstamp;
  }

  assert(!isnan(tstamp));

  assert(!isnan(lastRotUpdateTstamp));

  assert(!isnan(delta_time) || 
	 (fprintf(stderr, "DT %lf\n", delta_time) & 0));

  lastRotUpdateTstamp = tstamp;
  return delta_time;
}

void PoseRBFilter::UpdateGravity(double *readings, double tstamp,
				 double delta_t_rot) {
  double *m_body_accel = readings;

  for (int particlenum = 0; particlenum < particleCount; particlenum++) {
    RBParticle *particle = particles[particlenum];
    particle->TimeUpdateRotation(delta_t_rot);

    for (int i = 0; i < 3; i++)
      assert(!isnan(m_body_accel[i]));

    // calculate weight from measured gravity likelihood
    double grav_fixed_frame_estimate[3];
    particle->GetQuaternion()->RotateB2F(m_body_accel, grav_fixed_frame_estimate);
    for (int i = 0; i < 3; i++)
      assert(!isnan(grav_fixed_frame_estimate[i]));
    double log_grav_likelihood = gravMeasurePDF.LnValue(grav_fixed_frame_estimate);
      
    double log_particle_weight = particle->GetLogWeight();
    double new_log_weight = log_particle_weight + log_grav_likelihood;

    assert(!isnan(log_particle_weight));
    assert(!isnan(log_grav_likelihood));
    assert(!isnan(new_log_weight));

    particle->SetLogWeight(new_log_weight);
  }
}

void PoseRBFilter::UpdateGPSVel(double *readings, double tstamp,
				double delta_t_rot) {
  double *m_vel = readings;

  UpdatePosTstamp(tstamp);
  for (int particlenum = 0; particlenum < particleCount; particlenum++) {
    RBParticle *particle = particles[particlenum];
    particle->TimeUpdateRotation(delta_t_rot);

    kalmanGPSVelUpdater->Update(m_vel, 
				particle->GetPosMean(), 
				particle->GetPosCovar());

    // calculate weight from position measurement likelihood
    gpsVelMeasurePDF.SetMean(kalmanGPSVelUpdater->GetMeasurementLikelihoodMean());
    gpsVelMeasurePDF.SetCovar(kalmanGPSVelUpdater->GetMeasurementLikelihoodCovar());
    double log_vel_likelihood = gpsVelMeasurePDF.LnValue(m_vel);

    double log_particle_weight = particle->GetLogWeight();
    double new_log_weight = log_vel_likelihood + log_particle_weight; 
    assert(!isnan(new_log_weight));
    particle->SetLogWeight(new_log_weight);
  }
}

void PoseRBFilter::UpdateAcceleration(double *readings, double tstamp,
				      double delta_t_rot) {
  UpdatePosTstamp(tstamp);

  double *m_body_accel = readings;
  for (int particlenum = 0; particlenum < particleCount; particlenum++) {
    RBParticle *particle = particles[particlenum];
    particle->TimeUpdateRotation(delta_t_rot);

    // calculate residual acceleration, use to update Kalman filter  
    double residual_accel[3];
    CalcResidualAcceleration(particle, m_body_accel, residual_accel);
    double *rotationF2B = particle->GetQuaternion()->GetMatrixF2B();
    accelMeasurementMatrix.ReplaceSubmatrixA(rotationF2B, 3, 3);

    kalmanAccelUpdater->Update(residual_accel, 
			       particle->GetPosMean(), 
			       particle->GetPosCovar());

    // calculate weight from residual acceleration measurement likelihood
    accelMeasurePDF.SetMean(kalmanAccelUpdater->GetMeasurementLikelihoodMean());
    accelMeasurePDF.SetCovar(kalmanAccelUpdater->GetMeasurementLikelihoodCovar());
    double log_accel_likelihood = accelMeasurePDF.LnValue(residual_accel);

    for (int i = 0; i < 3; i++)
      assert(!isnan(m_body_accel[i]));

    // calculate weight from measured gravity likelihood
    double grav_fixed_frame_estimate[3];
    particle->GetQuaternion()->RotateB2F(m_body_accel, grav_fixed_frame_estimate);
    for (int i = 0; i < 3; i++)
      assert(!isnan(grav_fixed_frame_estimate[i]));
    double log_grav_likelihood = gravMeasurePDF.LnValue(grav_fixed_frame_estimate);
      
    double log_particle_weight = particle->GetLogWeight();
    double new_log_weight = log_particle_weight + log_grav_likelihood + log_accel_likelihood;

    assert(!isnan(log_particle_weight));
    assert(!isnan(log_grav_likelihood));
    assert(!isnan(new_log_weight));

    particle->SetLogWeight(new_log_weight);
  }
}

void PoseRBFilter::UpdateRateGyro(double *readings, double tstamp,
				  double delta_t_rot) {
  double *gyro = readings;
  
  for (int particlenum = 0; particlenum < particleCount; particlenum++) {
    RBParticle *particle = particles[particlenum];
    particle->TimeUpdateRotation(delta_t_rot);
    
    for (int i = 0; i < 3; i++)
      assert(!isnan(gyro[i]));
    
    particle->SetRotationalVelocity(gyro);
  }
}

void PoseRBFilter::UpdateMagnetometer(double *readings, double tstamp,
				      double delta_t_rot) {
  double *m_magfield_body = readings;

  for (int i = 0; i < 3; i++) {
    readings[i] -= magfieldBias[i];
  }

  for (int particlenum = 0; particlenum < particleCount; particlenum++) {
    RBParticle *particle = particles[particlenum];
    particle->TimeUpdateRotation(delta_t_rot);

    // calculate weight from measured magnetic field likelihood
    double magfield_fixed_frame_estimate[3];
    particle->GetQuaternion()->RotateB2F(m_magfield_body, magfield_fixed_frame_estimate);
    double log_magfield_likelihood = magfieldMeasurePDF.LnValue(magfield_fixed_frame_estimate);
      
    double log_particle_weight = particle->GetLogWeight();
    double new_log_weight = log_magfield_likelihood + log_particle_weight;
    assert(!isnan(new_log_weight));
    particle->SetLogWeight(new_log_weight);
  }
}

void PoseRBFilter::UpdateGPSPos(double *readings, double tstamp,
				double delta_t_rot) {
  // northing, easting, downing, sderr-xx, sderr-yy
  UpdatePosTstamp(tstamp); 

  double *pos_xyz = readings;

  for (int pos_i = 0; pos_i < 3; pos_i++) {
    pos_xyz[pos_i] -= gpsPosOffset[pos_i];
  }

  posMeasurementCovar.ReplaceGPSXX(readings[3]*readings[3]);
  posMeasurementCovar.ReplaceGPSYY(readings[4]*readings[4]);

    
  for (int particlenum = 0; particlenum < particleCount; particlenum++) {
    RBParticle *particle = particles[particlenum];
    particle->TimeUpdateRotation(delta_t_rot);

    kalmanPosUpdater->Update(pos_xyz, 
			     particle->GetPosMean(), particle->GetPosCovar());


    // calculate weight from position measurement likelihood
    posMeasurePDF.SetMean(kalmanPosUpdater->GetMeasurementLikelihoodMean());
    posMeasurePDF.SetCovar(kalmanPosUpdater->GetMeasurementLikelihoodCovar());
    double log_pos_likelihood = posMeasurePDF.LnValue(pos_xyz);

    double log_particle_weight = particle->GetLogWeight();
    double new_log_weight = log_pos_likelihood + log_particle_weight; 
    assert(!isnan(new_log_weight));

    particle->SetLogWeight(new_log_weight);
  }
}

void PoseRBFilter::UpdateBodyRelVelocity(double *readings, double tstamp,
					 double delta_t_rot, 
					 KalmanUpdater *kalmanUpdater) {
  UpdatePosTstamp(tstamp);

  double *m_vel = readings;
  for (int particlenum = 0; particlenum < particleCount; particlenum++) {
    RBParticle *particle = particles[particlenum];
    particle->TimeUpdateRotation(delta_t_rot);

    double *rotationF2B = particle->GetQuaternion()->GetMatrixF2B();
    velMeasurementMatrix.ReplaceSubmatrixV(rotationF2B, 3, 3);
    kalmanUpdater->Update(m_vel, 
			  particle->GetPosMean(), particle->GetPosCovar());
    // calculate weight from velocity measurement likelihood
    velMeasurePDF.SetMean(kalmanUpdater->GetMeasurementLikelihoodMean());
    velMeasurePDF.SetCovar(kalmanUpdater->GetMeasurementLikelihoodCovar());
    double log_vel_likelihood = velMeasurePDF.LnValue(m_vel);

    double log_particle_weight = particle->GetLogWeight();
    double new_log_weight =log_vel_likelihood + log_particle_weight; 
    assert(!isnan(new_log_weight));
    particle->SetLogWeight(new_log_weight);
  }
}

void PoseRBFilter::Update(double *readings, 
			  measurement_t measurement_type,
			  double tstamp) {
  double delta_t_rot = UpdateRotTstamp(tstamp);

  /* For each particle, update the rotation to the current time, 
     do a Kalman update if appropriate, and calculate measurement 
     likelihoods for importance weights.  Note that Kalman process 
     updates are not done for gyro/accelerometer updates for the sake
     of computational efficiency.
  */

  switch (measurement_type) {
  case R_GRAVONLY:
    UpdateGravity(readings, tstamp, delta_t_rot);
    break;
  case R_GPSVEL: 
    UpdateGPSVel(readings, tstamp, delta_t_rot);
    break;
  case R_ACCEL: 
    UpdateAcceleration(readings, tstamp, delta_t_rot);
    break;
  case R_GYRO: 
    UpdateRateGyro(readings, tstamp, delta_t_rot);
    break;
  case R_MAG:     
    UpdateMagnetometer(readings, tstamp, delta_t_rot);
    break;
  case R_GPSPOS: 
    UpdateGPSPos(readings, tstamp, delta_t_rot);
    break;
  case R_OD: 
    UpdateBodyRelVelocity(readings, tstamp, delta_t_rot, kalmanOdUpdater);
    break;
  case R_VISOD: 
    UpdateBodyRelVelocity(readings, tstamp, delta_t_rot, kalmanVisOdUpdater);
    break;
  default:
    fprintf(stderr, "Unknown measurement type '%d'\n", measurement_type);
    break;
  }  
}

RBParticle *PoseRBFilter::Estimate() {
  //  sort(particles, particles + particleCount, particle_comparer_descending());
  NormalizeParticleWeights();
  CalculateMeanParticle();
  return estimateParticle;
}

void PoseRBFilter::CalculateMeanParticle() {
  double sum_quatprods[4][4];
  double *expected_analytical_state = estimateParticle->GetPosMean();
  double mean_rotveloc[3];

  memset(&sum_quatprods[0][0], 0, sizeof(double)*16);
  memset(expected_analytical_state, 0, sizeof(double)*analyticalStateDim);
  memset(mean_rotveloc, 0, sizeof(double)*3);

  double sum_weights = 0.0;

  /* Mean of Gaussian is calculated as weighted mean of all Gaussians.

     Mean quaternion is calculated as the largest eigenvector of 
     the weighted sum of outer products of quaternions with themselves,
     as described in 'A note on mean orientation', A Morawiec, J. Applied 
     Crystallography, 1998
  */
  for (int particlenum = 0; particlenum < particleCount; particlenum++) {
    RBParticle *particle = particles[particlenum];
    double *quat = particle->GetQuaternion()->GetArray();
    double *analytical_state = particle->GetPosMean();
    double weight = particle->GetWeight();
    sum_weights += weight;

    cblas_dgemm(CblasRowMajor, CblasTrans, CblasNoTrans, 4, 4, 1,
                weight * 4.0, quat, 4, quat, 4, 1.0, &sum_quatprods[0][0], 4);

    for (int i = 0; i < 4; i++)
      sum_quatprods[i][i] -= weight;
    
    double *angveloc = particle->GetRotationalVelocity();
    for (int i = 0; i < 3; i++)
      mean_rotveloc[i] += angveloc[i];

    cblas_daxpy(analyticalStateDim, 
		weight, analytical_state, 1,
		expected_analytical_state, 1);
  }

  for (int i = 0; i < analyticalStateDim; i++)
    expected_analytical_state[i] /= sum_weights;

  for (int i = 0; i < 3; i++) 
    mean_rotveloc[i] /= particleCount;

  double eigenvals[4], eigenvecs[4][4];

  gsl_matrix_view gsl_quatprods = 
    gsl_matrix_view_array(&sum_quatprods[0][0], 4, 4);
  gsl_matrix_view gsl_eigenvecs = 
    gsl_matrix_view_array(&eigenvecs[0][0], 4, 4);
  gsl_vector_view gsl_eigenvals = gsl_vector_view_array(eigenvals, 4);
  
  gsl_eigen_symmv_workspace *gsl_eigen_temp = gsl_eigen_symmv_alloc(4);
  gsl_eigen_symmv(&gsl_quatprods.matrix, &gsl_eigenvals.vector,
                  &gsl_eigenvecs.matrix, gsl_eigen_temp);

  gsl_eigen_symmv_free(gsl_eigen_temp);

  gsl_eigen_symmv_sort(&gsl_eigenvals.vector, &gsl_eigenvecs.matrix,
                       GSL_EIGEN_SORT_ABS_DESC);
  
  double mean_quat[4];
  mean_quat[0] = eigenvecs[0][0];
  mean_quat[1] = eigenvecs[1][0];
  mean_quat[2] = eigenvecs[2][0];
  mean_quat[3] = eigenvecs[3][0];
  
  memcpy(estimateParticle->GetRotationalVelocity(), mean_rotveloc, 
	 sizeof(double)*3);
  memcpy(estimateParticle->GetQuaternion()->GetArray(), mean_quat, 
	 sizeof(double)*4);
  estimateParticle->GetQuaternion()->RecalcRotationMatrix();
}

double PoseRBFilter::EffectiveParticleCount() {
  double sum_squares = 0.0;

  NormalizeParticleWeights();

  for (int i = 0; i < particleCount; i++) {
    double weight = particles[i]->GetWeight();
    sum_squares += weight * weight;
  }

  return 1.0 / sum_squares;
}

RBParticle **PoseRBFilter::Particles() {
  sort(particles, particles + particleCount, particle_comparer_descending());
  return particles;
}

// Convert log weights to normal weights by expressing each as a 
// fraction in terms of a log reference weight, then normalizing 
// the fractions. 
double PoseRBFilter::NormalizeParticleWeights() {
  double ref_logweight = particles[0]->GetLogWeight();

  // ensures that weight ratio is always finite
  for (int i = 1; i < particleCount; i++) {
    if (particles[i]->GetLogWeight() > ref_logweight)
      ref_logweight = particles[i]->GetLogWeight();
  }

  assert(!isnan(ref_logweight));

  for (int i = 0; i < particleCount; i++) {
    double log_weight_ratio = particles[i]->GetLogWeight() - ref_logweight;
    assert(!isnan(particles[i]->GetLogWeight()));
    //    printf("LWR %lf\n", log_weight_ratio);
    double weight_ratio = exp(log_weight_ratio);
    if (isnan(weight_ratio)) {	
      weight_ratio = 0.0;
    }
    assert(!isnan(weight_ratio));
    assert(!isinf(weight_ratio));
    particles[i]->SetWeight(weight_ratio);
  }

  double sum_weights = 0.0;
  
  for (int i = 0; i < particleCount; i++) 
    sum_weights += particles[i]->GetWeight();
  
  for (int i = 0; i < particleCount; i++) {
    double normweight = particles[i]->GetWeight() / sum_weights;
    assert((!isnan(normweight)) ||
	   0*fprintf(stderr, "W = %.5e NW = %.5e\n", 
		     particles[i]->GetWeight(), normweight));
    particles[i]->SetWeight(normweight);
  }

  return sum_weights;
}

double PoseRBFilter::MaxWeight() {
  double maxweight = 0.0;

  for (int i = 0; i < particleCount; i++) {
    if (particles[i]->GetWeight() > maxweight)
      maxweight = particles[i]->GetWeight();
  }

  return maxweight;
}

void PoseRBFilter::CalcResidualAcceleration(RBParticle *particle,
					   double *accel, 
					   double *residual_out) {
  // Calculate expected gravity vector in body frame
  double gravity_body[3];
  particle->GetQuaternion()->RotateF2B(gravityRef, gravity_body);

  for (int i = 0; i < 3; i++)
    residual_out[i] = accel[i] - gravity_body[i];
 }
 
void PoseRBFilter::Resample() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  lastResampleTime = tv.tv_sec + 1e6 * tv.tv_usec;

  if (EffectiveParticleCount() >
      EFFECTIVE_PARTICLES_RESAMPLE_THRESHOLD * particleCount)
    return;
  else 
    fprintf(stderr, "R\n");

  NormalizeParticleWeights();

  // Particles don't need to be sorted for this to work 
  //  sort(particles.begin(), particles.end(), particle_comparer_ascending());
  double sum_weights = 0.0, cumulative_weights[particleCount],
    random_cumweights[particleCount];
  
  for (int i = 0; i < particleCount; i++) {
    assert(!isnan(particles[i]->GetWeight()));
    sum_weights += particles[i]->GetWeight();
    cumulative_weights[i] = sum_weights;
    //    fprintf(stderr, "weight(%d) = %lf\n", i, particles[i]->GetWeight());
  }

  assert(!isnan(sum_weights));
  
  for (int i = 0; i < particleCount; i++) 
    random_cumweights[i] = drand48() * sum_weights;
  
  sort(&random_cumweights[0], &random_cumweights[particleCount]);
  
  int new_i = 0;
  for (int old_i = 0; old_i < particleCount; old_i++) {
    double current_cum_weight = cumulative_weights[old_i];

    while (new_i < particleCount && 
           random_cumweights[new_i] <= current_cum_weight) {

      //      fprintf(stderr, "rcw[%d] = %lf\n", new_i, random_cumweights[new_i]);
      newParticles[new_i] = new RBParticle(*(particles[old_i]));
      newParticles[new_i]->SetWeight(1.0);
      newParticles[new_i]->SetLogWeight(0.0);
      new_i++;
    }

  }

  assert(new_i == particleCount);
  
  for (int i = 0; i < particleCount; i++) {
    delete particles[i];
  }

  RBParticle **old_particles = particles;
  particles = newParticles;
  newParticles = old_particles;
}

     


