/* 
   PoseRBFilter.h

   (C) 2005 Paul Vernaza, <vernaza@stwing.upenn.edu>
*/

#ifndef __IMU_RB_FILTER_H__
#define __IMU_RB_FILTER_H__

#define EFFECTIVE_PARTICLES_RESAMPLE_THRESHOLD (0.5)

#include <gsl/gsl_randist.h>
#include <gsl/gsl_rng.h>

#include "GaussianND.h"
#include "RBParticle.h"
#include "KalmanUpdater.h"
#include "ModelMatrices.h"

/* Types of measurements to pass to Update */
typedef enum { R_GRAVONLY,   // 3-dim accelerometer for pitch/roll correction
	       R_ACCEL,      // 3-dim accelerometer for pitch/roll correction
	                     //  and acceleration estimation
	       R_GYRO,       // 3-dim rate gyro
	       R_MAG,        // 3-dim magnetometer
	       R_OD,        // 3-dim body-relative velocity measurement
	       R_VISOD,     // 3-dim body-relative velocity measurement
	       R_GPSPOS,     // 5-dim GPS N'ing/E'ing/altitude/error x/error y
	       R_GPSVEL      // 3-dim GPS velocity
} measurement_t;

class PoseRBFilter {
 public:
  /* randomize_rotations: if true, rotations are randomly initialized */
  PoseRBFilter(int particle_count_, bool randomize_rotations = false);
  ~PoseRBFilter() {
    for (int i = 0; i < particleCount; i++) {
      delete particles[i];
    }
    delete kalmanPosUpdater;
    delete kalmanOdUpdater;
    delete kalmanVisOdUpdater;
    delete kalmanAccelUpdater;
    delete kalmanGPSVelUpdater;

    delete estimateParticle;

    free(particleBuffer1);
    free(particleBuffer2);    
  }

  /* Get mean particle of distribution */
  RBParticle *Estimate();

  /* Returns pointer to array of pointers to particles */
  RBParticle **Particles();

  /* Updates the filter with the data in readings of type measurement_type
     taken at time tstamp.  You must call this method with increasing 
     time stamps.
  */
  void Update(double *readings, measurement_t measurement_type, double tstamp);

  /* Resamples IF EffectiveParticleCount is below a threshold percentage */
  void Resample();

  /* Returns number of particles */
  int ParticleCount() { return particleCount; }

  /* Updates linear weights based on log weights.  NB: only log weights
     are updated with Update(), so YOU MUST CALL THIS BEFORE USING 
     LINEAR WEIGHTS.  Calling Update() again will invalidate the 
     linear weights.
  */
  double NormalizeParticleWeights();

  /* Returns time of last resampling */
  double LastResampleTime() { return lastResampleTime; }

  /* Returns a heuristic measure of about how many particles 
     have significant weights.
  */
  double EffectiveParticleCount();

  /* Returns the weight of the particle with maximum weight */
  double MaxWeight();

 protected:
  /* Pointer to current buffer of particles (array of RBParticle pointers) */
  RBParticle **particles;
  int particleCount;

 private:
  void CalcResidualAcceleration(RBParticle *particle,
				double *accel, double *residual_out);

  /* Calculates mean particle to update estimate     
   */
  void CalculateMeanParticle();

  /* Returns tstamp - tstamp of last call.
     Used to update rotations of particles.
   */
  double UpdateRotTstamp(double tstamp);

  /* Returns (tstamp - tstamp of last call), updates delta time parameters
     in Kalman matrices 
   */
  double UpdatePosTstamp(double tstamp);

  void UpdateGravity(double *readings, double tstamp, double delta_t_rot);
  void UpdateGPSVel(double *readings, double tstamp, double delta_t_rot);
  void UpdateAcceleration(double *readings, double tstamp, double delta_t_rot);
  void UpdateRateGyro(double *readings, double tstamp, double delta_t_rot);
  void UpdateMagnetometer(double *readings, double tstamp, double delta_t_rot);
  void UpdateGPSPos(double *readings, double tstamp, double delta_t_rot);
  void UpdateBodyRelVelocity(double *readings, double tstamp, double delta_t_rot, 
			     KalmanUpdater *kalmanUpdater);

  double gravityRef[3];
  double magfieldRef[3];
  double magfieldBias[3];
  double gpsPosOffset[3];

  double lastResampleTime;
  double lastRotUpdateTstamp, lastPosUpdateTstamp;

  /* Objects for handling matrices */
  ProcessMatrix processMatrix;
  ProcessCovar processCovar;
  PosMeasurementMatrix posMeasurementMatrix;
  VelMeasurementMatrix velMeasurementMatrix;
  GPSVelMeasurementMatrix gpsVelMeasurementMatrix;
  AccelMeasurementMatrix accelMeasurementMatrix;
  PosMeasurementCovar posMeasurementCovar;
  //  VelMeasurementCovar velMeasurementCovar;
  OdMeasurementCovar odMeasurementCovar;
  VisOdMeasurementCovar visOdMeasurementCovar;
  GPSVelMeasurementCovar gpsVelMeasurementCovar;
  AccelMeasurementCovar accelMeasurementCovar;
  GravMeasurementCovar gravMeasurementCovar;
  MagfieldMeasurementCovar magfieldMeasurementCovar;

  /* Objects for calculating Gaussian values */
  GaussianND accelMeasurePDF;
  GaussianND velMeasurePDF;
  //  GaussianND forwardVelMeasurePDF;
  GaussianND gpsVelMeasurePDF;
  GaussianND posMeasurePDF;
  GaussianND gravMeasurePDF;
  GaussianND magfieldMeasurePDF;

  /* Objects for calculating Kalman updates */
  KalmanUpdater *kalmanPosUpdater, *kalmanAccelUpdater,
    *kalmanGPSVelUpdater, *kalmanOdUpdater, *kalmanVisOdUpdater;

  int analyticalStateDim;

  /* Buffers to store references to particle objects */
  RBParticle **particleBuffer1, **particleBuffer2, **newParticles;
  
  /* Stores current mean estimate */
  RBParticle *estimateParticle;
};

#endif
