/* 
   Quaternion.cpp

   (C) 2005 Paul Vernaza, <vernaza@stwing.upenn.edu>
*/

#include "Quaternion.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include <gsl/gsl_cblas.h>

// Note that quaternion real part here is the LAST component 

Quaternion::Quaternion() {
  rotationMatrixDirty = true;
  quaternion = &quaternion_storage[0];
  quaternion[0] = 0.0;
  quaternion[1] = 0.0;
  quaternion[2] = 0.0;
  quaternion[3] = 1.0;
  rotationCount = 0;
}

/*
  Initialize from specified quaternion components (w is real part)
*/
Quaternion::Quaternion(double i, double j, double k, double w) {
  rotationMatrixDirty = true;
  quaternion = &quaternion_storage[0];
  quaternion[0] = i;
  quaternion[1] = j;
  quaternion[2] = k;
  quaternion[3] = w;
  rotationCount = 0;
}

/*
  Initialize from Euler angles
*/
Quaternion::Quaternion(double roll, double pitch, double yaw) {
  rotationMatrixDirty = true;
  quaternion = &quaternion_storage[0];
  quaternion[0] = 0.0;
  quaternion[1] = 0.0;
  quaternion[2] = 0.0;
  quaternion[3] = 1.0;

  rotationCount = 0;
  double rotation[3];
  rotation[0] = rotation[1] = rotation[2] = 0.0;

  rotation[0] = roll; rotation[1] = 0.0; rotation[2] = 0.0;
  IntegrateRotation(1.0, rotation);
  rotation[0] = 0.0; rotation[1] = pitch; rotation[2] = 0.0;
  IntegrateRotation(1.0, rotation);
  rotation[0] = 0.0; rotation[1] = 0.0; rotation[2] = yaw; 
  IntegrateRotation(1.0, rotation);
  
  /*
  double phi = roll/2.0, theta = pitch/2.0 , psi = yaw/2.0;
  quaternion[0] = cos(phi)*cos(theta)*cos(psi) + sin(phi)*sin(theta)*sin(psi);
  quaternion[1] = sin(phi)*cos(theta)*cos(psi) - cos(phi)*sin(theta)*sin(psi);
  quaternion[2] = cos(phi)*sin(theta)*cos(psi) + sin(phi)*cos(theta)*sin(psi);
  quaternion[3] = cos(phi)*cos(theta)*cos(psi) - sin(phi)*sin(theta)*sin(psi);
  */
}

/*
Quaternion::Quaternion(double *quat_ptr) {
  rotationMatrixDirty = true;
  quaternion = quat_ptr;
}
*/

Quaternion::Quaternion(Quaternion &quat) {
  quaternion = &quaternion_storage[0];
  memcpy(quaternion, quat.GetArray(), sizeof(double) * 4);
  rotationMatrixDirty = true;
}

/*
  Provides Euler angles representing this rotation.  angles should be 
  a pre-allocated double[4]
*/
void Quaternion::GetEulerAngles(double *angles) {
  GetMatrixB2F();
  //  GetMatrixF2B();

  double yaw = atan2(-rotationMatrixB2F[0][1], rotationMatrixB2F[0][0]);
  double roll = atan2(-rotationMatrixB2F[1][2], rotationMatrixB2F[2][2]);
  double pitch = atan2(rotationMatrixB2F[0][2], 
		       rotationMatrixB2F[2][2] / (cos(roll) + 0.000000001));

  angles[0] = roll;
  angles[1] = pitch;
  angles[2] = yaw;
}

/*
void Quaternion::SetArray(double *quat) {
  memcpy(quaternion, quat, sizeof(double) * 4);
  rotationMatrixDirty = true;
}
*/

/*
  Returns a pointer to a 3x3 matrix that transforms vectors in the body 
  frame to the fixed frame.  Don't modify the referenced matrix.
*/
double *Quaternion::GetMatrixB2F() {
  if (!rotationMatrixDirty) return &rotationMatrixB2F[0][0];

  for (int i = 0; i < 4; i++)
    assert(!isnan(quaternion[i]));

  rotationMatrixDirty = false;

  double b1 = quaternion[0];
  double b2 = quaternion[1];
  double b3 = quaternion[2];
  double b4 = quaternion[3];

  double p1 = b1 * b1;
  double p2 = b2 * b2;
  double p3 = b3 * b3;
  double p4 = b4 * b4;
  double p5 = p2 + p3;

  double p6 = p5 + p4 + p1;
  if (p6 != 0) p6 = 2.0 / p6;

  rotationMatrixB2F[0][0] = 1 - p6 * p5;
  rotationMatrixB2F[1][1] = 1 - p6 * (p1 + p3);
  rotationMatrixB2F[2][2] = 1 - p6 * (p1 + p2);
  
  p1 = p6 * b1;
  p2 = p6 * b2;
  p5 = p6 * b3 * b4;
  p6 = p1 * b2;

  rotationMatrixB2F[0][1] = p6 - p5;
  rotationMatrixB2F[1][0] = p6 + p5;
  
  p5 = p2 * b4;
  p6 = p1 * b3;

  rotationMatrixB2F[0][2] = p6 + p5;
  rotationMatrixB2F[2][0] = p6 - p5;

  p5 = p1 * b4;
  p6 = p2 * b3;

  rotationMatrixB2F[1][2] = p6 - p5;
  rotationMatrixB2F[2][1] = p6 + p5;

  //  PrintMatrix(&rotationMatrixB2F[0][0]);


  /*
  double tempmat[3][3];
  memcpy(&tempmat[0][0], &rotationMatrixB2F[0][0], sizeof(double)*9);
  for (int row = 0; row < 3; row++)
    for (int col = 0; col < 3; col++)
      rotationMatrixB2F[col][row] = tempmat[row][col];
  */


  return &rotationMatrixB2F[0][0];
}

/*
  Returns a pointer to a 3x3 matrix that transforms vectors in the fixed 
  frame to the body frame.  Don't modify the referenced matrix.
*/
double *Quaternion::GetMatrixF2B() {
  GetMatrixB2F();

  memcpy(rotationMatrixF2B, rotationMatrixB2F, sizeof(double)*9);
  for (int row = 0; row < 3; row++)
    for (int col = 0; col < 3; col++)
      rotationMatrixF2B[col][row] = rotationMatrixB2F[row][col];
  return &rotationMatrixF2B[0][0];
}

/*
  Makes this a unit quaternion by normalizing the components
*/
void Quaternion::Renormalize() {
  double norm_orientation_quat = cblas_dnrm2(4, quaternion, 1);
  double normalization_factor = 1.0 / norm_orientation_quat;
  quaternion[0] *= normalization_factor;
  quaternion[1] *= normalization_factor;
  quaternion[2] *= normalization_factor;
  quaternion[3] *= normalization_factor;
}

/*
  Conjugates this quaternion
*/
void Quaternion::Conjugate() {
  quaternion[0] *= -1;
  quaternion[1] *= -1;
  quaternion[2] *= -1;
}

/*
  Ensures that the real part of this quaternion is positive without 
  changing the rotation
*/
void Quaternion::MakePositive() {
  if (quaternion[3] < 0.0) 
    for (int i = 0; i < 4; i++) quaternion[i] *= -1;
}

void Quaternion::Invert() {
  double normsquare = 
    quaternion[0] * quaternion[0] + 
    quaternion[1] * quaternion[1] + 
    quaternion[2] * quaternion[2] + 
    quaternion[3] * quaternion[3];

  Conjugate();
  
  for (int i = 0; i < 4; i++) quaternion[i] /= normsquare;
}

/*
  Multiplies this quaternion by otherquat on the right
*/
void Quaternion::RMultiply(Quaternion *otherquat) {
  double *otherquat_array = otherquat->GetArray();
  double a = quaternion[3], b = quaternion[0], c = quaternion[1],
    d = quaternion[2];
  double t = otherquat_array[3], x = otherquat_array[0], 
    y = otherquat_array[1], z = otherquat_array[2];

  quaternion[3] = a*t - b*x - c*y - d*z;
  quaternion[0] = b*t + a*x + c*z - d*y;
  quaternion[1] = c*t + a*y + d*x - b*z;
  quaternion[2] = d*t + z*a + b*y - x*c;
}

/*
  Rotates this quaternion for delta_time about the axes of the body frame
  with angular rates specified by three rates in angular_rates
*/
void Quaternion::IntegrateRotation(double delta_time, 
                                   double *angular_rates) {
  double velocity_norm = delta_time * cblas_dnrm2(3, angular_rates, 1);

  if (isinf(fabs(velocity_norm))) {
    fprintf(stderr, "WARNING: Quaternion::IntegrateRotation(): corrupt velocity data?\n");
    return;
  }

  if (velocity_norm == 0.0) return;

  for (int i = 0; i < 3; i++) 
    assert(!isnan(angular_rates[i]));

  assert(!isnan(delta_time));

  assert(!isnan(velocity_norm));

  //  fprintf(stderr, "D %f V %f\n", delta_time, velocity_norm);

  double cos_part = cos(velocity_norm / 2.0);
  double sin_part = 
    delta_time * sin(velocity_norm / 2.0) / velocity_norm;

  assert(!isnan(sin_part) || (printf("VN %lf DT %lf AR %lf %lf %lf\n", velocity_norm, delta_time, angular_rates[0], angular_rates[1], angular_rates[2]) && 0));

  double p = sin_part * angular_rates[0];
  double q = sin_part * angular_rates[1];
  double r = sin_part * angular_rates[2];
  
  double differentialQuaternionMatrix[4][4];
  differentialQuaternionMatrix[0][0] = cos_part;
  differentialQuaternionMatrix[0][1] = r;
  differentialQuaternionMatrix[0][2] = -q;
  differentialQuaternionMatrix[0][3] = p;

  differentialQuaternionMatrix[1][0] = -r;
  differentialQuaternionMatrix[1][1] = cos_part;
  differentialQuaternionMatrix[1][2] = p;
  differentialQuaternionMatrix[1][3] = q;
  
  differentialQuaternionMatrix[2][0] = q;
  differentialQuaternionMatrix[2][1] = -p;
  differentialQuaternionMatrix[2][2] = cos_part;
  differentialQuaternionMatrix[2][3] = r;

  differentialQuaternionMatrix[3][0] = -p;
  differentialQuaternionMatrix[3][1] = -q;
  differentialQuaternionMatrix[3][2] = -r;
  differentialQuaternionMatrix[3][3] = cos_part;

  double result[4];

  /*
  cblas_dgemv(CblasRowMajor, CblasNoTrans, 4, 4,
              1.0, &differentialQuaternionMatrix[0][0], 4,
              &quaternion[0], 1,
              0.0, &quaternion[0], 1);
  */
  cblas_dgemv(CblasRowMajor, CblasNoTrans, 4, 4,
              1.0, &differentialQuaternionMatrix[0][0], 4,
              &quaternion[0], 1,
              0.0, &result[0], 1);
  memcpy(quaternion, result, sizeof(double) * 4);

  // norm tends to shrink over time...
  if ((rotationCount++ % 100) == 0)
    Renormalize();
  /*
  fprintf(stderr, "QR %f %f %f %f\n", 
         quaternion[0],
         quaternion[1],
         quaternion[2],
         quaternion[3]);
  */

  //fprintf(stderr, "QR %f\n", cblas_dnrm2(3, quaternion, 1));

  rotationMatrixDirty = true;
}

/*
  Convert vector in body frame to fixed frame.  fixed_vector_out 
  must point to allocated array of three doubles.
*/
void Quaternion::RotateB2F(double *body_vector, 
			   double *fixed_vector_out) {
  double *rotationMatrixPtr = GetMatrixB2F();
  cblas_dgemv(CblasRowMajor, CblasNoTrans, 3, 3, 
              1.0, rotationMatrixPtr, 3, body_vector, 1, 
              0.0, fixed_vector_out, 1);
  /*
  Quaternion vecquat;
  double *vecquat_array = vecquat.GetArray();
  for (int i = 0; i < 3; i++) vecquat_array[i] = body_vector[i];
  vecquat_array[3] = 0.0;

  Quaternion tempquat(*this);
  Quaternion tempquat_inv(tempquat);
  tempquat_inv.Invert();

  tempquat.RMultiply(&vecquat);
  tempquat.RMultiply(&tempquat_inv);
  memcpy(fixed_vector_out, tempquat.GetArray(), sizeof(double) * 3);
  */
}

/*
  Convert vector in fixed frame to body frame.  body_vector_out 
  must point to allocated array of three doubles.
*/
void Quaternion::RotateF2B(double *fixed_vector, 
			   double *body_vector_out) {
  double *rotationMatrixPtr = GetMatrixB2F();
  cblas_dgemv(CblasRowMajor, CblasTrans, 3, 3, 
              1.0, rotationMatrixPtr, 3, fixed_vector, 1, 
              0.0, body_vector_out, 1);
  /*
  Quaternion vecquat;
  double *vecquat_array = vecquat.GetArray();
  for (int i = 0; i < 3; i++) vecquat_array[i] = fixed_vector[i];
  vecquat_array[3] = 0.0;

  Quaternion tempquat(*this);
  Quaternion tempquat_inv(tempquat);
  tempquat_inv.Invert();

  tempquat_inv.RMultiply(&vecquat);
  tempquat_inv.RMultiply(&tempquat);
  memcpy(body_vector_out, tempquat_inv.GetArray(), sizeof(double) * 3);  
  */
}

Quaternion *Quaternion::RDivide(Quaternion *rquat) {
  Quaternion inv_rquat(*rquat);
  inv_rquat.Conjugate();
  Quaternion *result = new Quaternion(*this);
  result->RMultiply(&inv_rquat);
  return result;
}

double Quaternion::GetAngleDiff(Quaternion *quat) {
  MakePositive();
  quat->MakePositive();

  Quaternion *div = RDivide(quat);
  div->MakePositive();
  double *quatarray = div->GetArray();
  double angle = 2 * acos(quatarray[3]);
  if (angle >= M_PI)
    angle -= M_PI;
  return angle;
}

void Quaternion::Print(FILE *file) {
  fprintf(file, "Q %f %f %f %f\n", 
         quaternion[0], quaternion[1], quaternion[2], quaternion[3]);
}

/*
void Quaternion::PrintMatrix(double *matrix) {
  fprintf(stderr, "\n[ %f %f %f ]", matrix[0], matrix[1], matrix[2]);
  fprintf(stderr, "\n[ %f %f %f ]", matrix[3], matrix[4], matrix[5]);
  fprintf(stderr, "\n[ %f %f %f ]\n", matrix[6], matrix[7], matrix[8]);
}
*/

#ifdef TESTSCRIPT

int main(int argc, char *argv[]) {
  printf("bar\n");
}

#endif

