/* 
   RBParticle.h

   (C) 2005 Paul Vernaza, <vernaza@stwing.upenn.edu>
*/

#ifndef __RBPARTICLE_H__
#define __RBPARTICLE_H__

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <gsl/gsl_randist.h>

#include "Quaternion.h"
#include "SystemParameters.h"
#include "DebugMatrices.h"

class RBParticle {
 public:
  //  RBParticle();
  RBParticle(int pos_dim);
  RBParticle(int pos_dim, double *initial_mean, double *initial_covar);
  RBParticle(RBParticle &src);
  ~RBParticle() {
    gsl_rng_free(prng);
    free(posMean);
    free(posCovar);
    delete quaternion;
  };

  /* Randomize orientation */
  void Randomize();

  /* Get/set the log weight */
  double GetLogWeight() { return logWeight; }
  void SetLogWeight(double weight_arg) { logWeight = weight_arg; }

  /* Get/set the linear weight */
  double GetWeight() { return weight; }
  void SetWeight(double weight_arg);

  /* Zero out position state */
  void ZeroPosition() { memset(posMean, 0, sizeof(double)*posDim); }

  /* Get mean, covariance of position state Gaussian */
  double *GetPosMean() { return posMean; }
  double *GetPosCovar() { return posCovar; }

  /* Return a pointer to the internal quaternion */
  Quaternion *GetQuaternion() { return quaternion; }

  /* Replaces the internal quaternion with the one pointed to by quat */
  void SetNewQuaternion(Quaternion *quat) {
    delete quaternion;
    quaternion = quat;
  }

  /* Set the rotational velocity of this particle, adding Gaussian noise */
  void SetRotationalVelocity(double *rot_in) {
    for (int i = 0; i < 3; i++) {
      double gauss_err = gsl_ran_gaussian(prng, Angular_Rate_Measurement_Error);
      rotVeloc[i] = rot_in[i] + gauss_err;
      assert(!isnan(rotVeloc[i]));
    }
  }

  /* Advance the rotation for the given time */
  void TimeUpdateRotation(double delta_t) {
    assert(!isnan(delta_t));

    quaternion->IntegrateRotation(delta_t, rotVeloc);
  }

  /* Returns pointer to 3d body-relative rotational velocity */
  double *GetRotationalVelocity() { return rotVeloc; }

 private:
  Quaternion *quaternion;
  double *posMean, *posCovar;
  double rotVeloc[3];
  double weight, logWeight;
  gsl_rng *prng;
  int posDim;

  void AllocateMatrices(int posDim);
  void PrintMatrix(double *matrix);

};

#endif
