#!/usr/bin/perl

=pod
   gen_matrices.pl

   (C) 2005 Paul Vernaza, <vernaza@stwing.upenn.edu>
=cut

use warnings; use strict;
use diagnostics;

my $Matrix_Name_Regex = qr/^N:([\w]+)\s*$/;

my ($infile, $outfile) = @ARGV;

die "usage: gen_matrices.pl <description file> <output file>" 
    unless $infile && $outfile;

open INFILE, $infile || die "open: $!";

my $state = 'matrix name';

my %matrices;

my $line;
my $current_matrix;
while ($line = <INFILE>) {
    next if $line =~ /^\s*#/;
    next if $line =~ /^\s*$/;

    chomp $line;

    if ($state eq 'matrix name') {
	($current_matrix) = ($line =~ /$Matrix_Name_Regex/);
#	print "Generating $current_matrix.h\n";
	$state = 'matrix';
	$matrices{$current_matrix} = {};
	$matrices{$current_matrix}->{lines} = [];
    } elsif ($state eq 'matrix') {
	if ($line =~ /$Matrix_Name_Regex/) {
	    $state = 'matrix name';
	    redo;
	}
	push @{$matrices{$current_matrix}->{lines}}, $line;
    }
}

for my $matrix_name (keys %matrices) {
    my $matrix = $matrices{$matrix_name};
    $matrix->{rows} = scalar(@{$matrix->{lines}});
 
    my $row = 0;
    for my $line (@{$matrix->{lines}}) {
	 my $col = 0;
	my @row = split /\t/, $line;
	$matrix->{cols} = scalar(@row);
	for (my $i = 0; $i < @row; $i++) {
	    $matrix->{array}[$row][$i] = $row[$i];
	}
	$row++;
    }
}

open FH, ">$outfile" || die "open: $!";
print_header();

for my $matrix_name (keys %matrices) {
    my $matrix = $matrices{$matrix_name};
    my $rows = $matrix->{rows};
    my $cols = $matrix->{cols};

    print_constructor_destructor($matrix, $matrix_name);
    print_tag_methods($matrix);
    print_submatrix_methods($matrix);
    print_trailer();
}
    
print FH<<END;
#endif
END
close FH;

sub print_tag_methods {
    my ($matrix) = @_;

    for my $tag (keys %{$matrix->{tags}}) {
	print FH<<END;
    void Replace$tag(double tagval) {
END
        for my $tag_loc (@{$matrix->{tags}{$tag}}) {
	    my ($tagrow, $tagcol) = @$tag_loc;
            print FH<<END;
	tempMatrix[$tagrow * cols + $tagcol] = tagval;
END
        }
        print FH<<END;
    }

END
    }

    for my $tag (keys %{$matrix->{multags}}) {
	print FH<<END;
    void Replace$tag(double tagval) {
END
        for my $tag_loc (@{$matrix->{multags}{$tag}}) {
	    my ($tagrow, $tagcol) = @$tag_loc;
            print FH<<END;
	tempMatrix[$tagrow * cols + $tagcol] = templateMatrix[$tagrow * cols + $tagcol] * tagval;
END
        }
        print FH<<END;
    }

END
    }
}

sub print_submatrix_methods {
    my ($matrix) = @_;

    for my $submatrix_name (keys %{$matrix->{submatrices}}) {
	my ($subrow, $subcol) = 
	    @{$matrix->{submatrices}{$submatrix_name}};
	print FH<<END;
    double *GetSubmatrix$submatrix_name(double tagval) {
	return tempMatrix + $subrow * cols + $subcol;
    }

    void ReplaceSubmatrix$submatrix_name(double *src, int srcrows, int srccols) {
	CopySubmatrix(src, srcrows, srccols, $subrow, $subcol);
    }

END
	
    }
}

sub print_constructor_destructor {
    my ($matrix, $matrix_name) = @_;
    my $rows = $matrix->{rows};
    my $cols = $matrix->{cols};

    print FH<<END;
class $matrix_name : public AutoMatrix {
  public:
    $matrix_name() {
	rows = $rows;
	cols = $cols;
	tempMatrix = (double*)calloc(rows*cols, sizeof(double));
	templateMatrix = (double*)calloc(rows*cols, sizeof(double));
END

    for (my $row = 0; $row < $rows; $row++) {
	for (my $col = 0; $col < $cols; $col++) {
	    my @tokens = split / *\* */, $matrix->{array}[$row][$col];
	    my $template_val = 0.0;

	    if ($tokens[0] =~ /^[[:alpha:]]/) {	
		# submatrix
		if (my ($subname) = ($tokens[0] =~ /^SUB_(\w+)/)) {
		    if (!defined($matrix->{submatrices}{$subname})) {
			$matrix->{submatrices}{$subname} = [ $row, $col ];
		    }
		} else { # tag for replacement 
		    if (!defined($matrix->{tags}{$tokens[0]})) {
			$matrix->{tags}{$tokens[0]} = [];
		    }
		    push @{$matrix->{tags}{$tokens[0]}}, [ $row, $col ];
		       }
	    } else {
		$template_val = $tokens[0];
	    }

	    if ($tokens[1]) {	# must be multiplication tag
		if (!defined($matrix->{multags}{$tokens[1]})) {
		    $matrix->{multags}{$tokens[1]} = [];
		}
		push @{$matrix->{multags}{$tokens[1]}}, [ $row, $col ];
	    }

	    my $template_val_str = sprintf "%.20e", $template_val;

	    print FH<<END;
         templateMatrix[$row * cols + $col] = $template_val_str;
END
	}
    }

    print FH<<END;
	memcpy(tempMatrix, templateMatrix, sizeof(double)*rows*cols);
    }

    ~$matrix_name() {
	free(tempMatrix);
	free(templateMatrix);
    }

END

}

sub print_header {
    print FH<<END;
#ifndef __MODEL_MATRICES_H__
#define __MODEL_MATRICES_H__
#include <stdlib.h>
#include <string.h>
#include "AutoMatrix.h"

END
}

sub print_trailer {
    print FH<<END;
};

END
}

