/*
  lagr/LagrPoseFilter.cpp

  (C) 2005 Paul Vernaza, <vernaza@stwing.upenn.edu>
*/

#include "LagrPoseFilter.h"

#include <arpa/inet.h>
#include <lagrInterface.h>
#include <netinet/in.h>
#include <gsl/gsl_cblas.h>
#include <string.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/types.h>

#include "lagr_penn.h"
#include "../UTM/LatLong-UTMconversion.h"

LagrPoseFilter::LagrPoseFilter(int particlecount, double message_lag) :
  PoseRBFilter(particlecount),
  msorter(message_lag),
  forwardPose(false)
{
  fprintf(stderr, "lag = %lf\n", message_lag);

  memset(gyroBias, 0, sizeof(double)*3);
  poseState = S_FirstUpdate;
  biasReadingCount = 0;
  memset(sumGyroReadings, 0, sizeof(double)*3);
  memset(estimatedBias, 0, sizeof(double)*3);

  exportPose = false;

  exportDisplay = false;

  messageLag = message_lag;

  Randomize();
}

void LagrPoseFilter::Randomize() {
  // randomize initial rotations
  double rotvel[3];
  rotvel[0] = 0.0; rotvel[1] = 0.0; 
  for (int i = 0; i < particleCount; i++) {
    rotvel[2] = ((double)i / particleCount) * 2.0 * M_PI;
    particles[i]->GetQuaternion()->IntegrateRotation(1.0, rotvel);
  }
}

void LagrPoseFilter::AddLagrMessage(LagrMessageType type, unsigned char *data) {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  double robot_time = tv.tv_sec + 1e-6 * tv.tv_usec;
  
  if (poseState == S_FirstUpdate) {
    startTime = robot_time;
    poseState = S_PoseUninitialized;
  }

  LagrMessageWrapper *mwrap;

  if (poseState == S_PoseUninitialized) {
    if (robot_time - startTime > BIAS_ESTIMATION_TIMELIMIT) {
      for (int i = 0; i < 3; i++) 
	estimatedBias[i] = sumGyroReadings[i] / biasReadingCount;
      if (cblas_dnrm2(3, estimatedBias, 1) > BIAS_MAGNITUDE_LIMIT) 
	memset(estimatedBias, 0, sizeof(double)*3);
      
      printf("Estimated bias: [ %.5e %.5e %.5e ]\n", 
	     estimatedBias[0], estimatedBias[1], estimatedBias[2]);
      
      poseState = S_GrabBaselinePose;
    }
  }

  double timestamp;
  switch (type) {
    // visual odometry
  case lagrMessageGeneric: {
    struct lagr_penn_header *header = (struct lagr_penn_header*)data;
    if(header->magicNumber != 0x50454e30) {
      fprintf(stderr, "LagrPoseFilter: bad magic number!\n");
      return;
    }
    if(header->messageType != lagrPennMessageVisOdometry) {
      fprintf(stderr, "LagrPoseFilter: expecting visual odometry message!\n");
      return;
    }
    unsigned char *payload_start = 
      (unsigned char*)header + sizeof(struct lagr_penn_header);

    mwrap = new LagrMessageWrapper(type, payload_start, sizeof(double));

    if (!msorter.EnqueueMessage(header->timeStamp, mwrap))
      delete mwrap;
  }
    break;
  case lagrMessageEncoder: {
    //    printf("encoder\n");

    LagrEncoderMessage *encoder_message = (LagrEncoderMessage*)data;
    
    LagrEncoder *encoder_start = &(encoder_message->samplesP[0]);
    LagrEncoder *encoder_end = 
      &(encoder_message->samplesP[encoder_message->count - 1]);
    timestamp = (encoder_start->timeStamp + encoder_end->timeStamp) / 2.0;
    //      timestamp = encoder_start->timeStamp;
    
    // convert absolute encoder to encoder difference
    int encdiff[2] = { encoder_end->left - encoder_start->left, 
		       encoder_end->right - encoder_start->right };
    double tdiff = encoder_end->timeStamp - encoder_start->timeStamp;
    
    for (int enci = 0; enci < 2; enci++) {
      if (fabs(encdiff[enci]) > MAX_ENCODER_JUMP)
	encdiff[enci] = 0;
    }
      
    encoder_start->timeStamp = tdiff;
    encoder_start->left = encdiff[0];
    encoder_start->right = encdiff[1];
      
    mwrap = new LagrMessageWrapper(type, (unsigned char*)encoder_start, sizeof(LagrEncoder));
      
    if (!msorter.EnqueueMessage(timestamp, mwrap))
      delete mwrap;
  }
    break;
  case lagrMessageIMU: {
    LagrIMUMessage *imu_message = (LagrIMUMessage*)data;

    for (int i = 0; i < imu_message->count; i++) {
      LagrIMU *imu_state = &(imu_message->samplesP[i]);
      timestamp = imu_state->timeStamp;

      if (poseState == S_PoseUninitialized) {
	sumGyroReadings[0] += imu_state->xRate;
	sumGyroReadings[1] += imu_state->yRate;
	sumGyroReadings[2] += imu_state->zRate;
	biasReadingCount++;
      }

      assert(!isnan(timestamp));

      mwrap = new LagrMessageWrapper(type, (unsigned char *)imu_state, sizeof(LagrIMU));
      if (!msorter.EnqueueMessage(timestamp, mwrap))
	delete mwrap;
    }
  }
    break;
  case lagrMessagePosition: {
    LagrPoseState *poseP = (LagrPoseState*)data;
    timestamp = poseP->timeStamp; 

    assert(!isnan(timestamp));

    if (poseState == S_GrabBaselinePose && poseP->estimateIsInitialized) {
      InitFilter(poseP);
      poseState = S_Run;
    }

    mwrap = new LagrMessageWrapper(type, data, sizeof(LagrPoseState));
    if (!msorter.EnqueueMessage(timestamp, mwrap))
      delete mwrap;
  }
    break;
  case lagrMessageGPS: {
    LagrGPS *gpsP = (LagrGPS*)data;
    timestamp = gpsP->timeStamp;

    if (gpsP->fix < 2)		// need at least a 2D fix for this to be valid
      break;

    // negative sign: north-east-down coordinates
    gpsP->mslHeight = - gpsP->mslHeight;
    gpsP->velUp = - gpsP->velUp;

    // NB: converted to UTM and stuffed back inside as latitude and longitude
    char utmzone[4];
    double deg_lat = 180 * gpsP->latitude / M_PI;
    double deg_long = 180 * gpsP->longitude / M_PI;
    double northing, easting;
    LLtoUTM(23, deg_lat, deg_long, northing, easting, utmzone);

    double xyz_vel[6] = { northing, 
			  easting,
			  gpsP->mslHeight,
			  gpsP->velNorth,
			  gpsP->velEast,
			  gpsP->velUp };

    gpsFixer.ProjectGPSForward(timestamp, 
			       gpsFixer.LastTimestamp(), 
			       xyz_vel);

    gpsP->latitude = xyz_vel[0];
    gpsP->longitude = xyz_vel[1];
    gpsP->mslHeight = xyz_vel[2];
    gpsP->velNorth = xyz_vel[3];
    gpsP->velEast = xyz_vel[4];
    gpsP->velUp = xyz_vel[5];

    assert(!isnan(timestamp));

    mwrap = new LagrMessageWrapper(type, data, sizeof(LagrGPS));
    if (!msorter.EnqueueMessage(gpsFixer.LastTimestamp(), mwrap))
      delete mwrap;
  }
    break;
  default:
    break;
  }

  double mesststamp;
  while (msorter.GetMessage(&mwrap, &mesststamp)) {
    ProcessMessage(mesststamp, mwrap);
    delete mwrap;
  }
}

void LagrPoseFilter::ProcessMessage(double timestamp, LagrMessageWrapper *mwrap) {
  LagrMessageType mtype = mwrap->type;
  unsigned char *mdata = mwrap->data;

  switch (mtype) {
  case lagrMessageGeneric: { // visual odometry 
    double bodyrelvel[3] = { 0.0, 0.0, 0.0 };
    bodyrelvel[0] = ((double*)mdata)[0];

    fprintf(stderr, "Vod: %lf\n", bodyrelvel[0]);

    Update(bodyrelvel, R_VISOD, timestamp);
  }
    break;
  case lagrMessageEncoder: {
    LagrEncoder *encoder_data = (LagrEncoder*)mdata;
    double tdiff = encoder_data->timeStamp;
    int encleft = encoder_data->left, encright = encoder_data->right;

    double wheelvelL = 
      encleft * ENCODER_RADIANS_PER_TICK * WHEEL_RADIUS / tdiff;
    double wheelvelR = 
      encright * ENCODER_RADIANS_PER_TICK * WHEEL_RADIUS / tdiff;

    double bodyrelvel[3];
    bodyrelvel[2] = 0.0;

    if (fabs(wheelvelL - wheelvelR) < 0.0001) {
      bodyrelvel[0] = wheelvelL; bodyrelvel[1] = 0.0;
    } else {
      double turncenter = (FRONT_AXLE_LEN / 2.0) * 
	(wheelvelL + wheelvelR) / (wheelvelL - wheelvelR);

      double theta = 
	wheelvelR / ((FRONT_AXLE_LEN / 2.0) - turncenter + 0.000000001);

      bodyrelvel[0] = - turncenter * theta;
      bodyrelvel[1] = 0.0;

      //      printf("Tcent %.5e\n", turncenter);
    }

    //    printf("V %lf %lf\n", bodyrelvel[0], bodyrelvel[1]);

    Update(bodyrelvel, R_OD, timestamp);

    Resample();

    gpsFixer.AddEstimate(timestamp, Estimate()->GetPosMean());
  }
    break;
  case lagrMessageIMU: {
    LagrIMU *imuP = (LagrIMU*)mdata;
    double gyro[3];
    double accel[3];
    gyro[0] = imuP->xRate - estimatedBias[0];
    gyro[1] = imuP->yRate - estimatedBias[1];
    gyro[2] = imuP->zRate - estimatedBias[2];
    accel[0] = imuP->xAcceleration;
    accel[1] = - imuP->yAcceleration; // COMPENSATE FOR INCONSISTENT FRAMES
    accel[2] = - imuP->zAcceleration;

    assert(!isnan(imuP->xRate) && !isnan(imuP->yRate) && !isnan(imuP->zRate));

    assert(!isnan(estimatedBias[0]) && !isnan(estimatedBias[1]) && !isnan(estimatedBias[2]));

    assert(!isnan(timestamp) || 
	   (fprintf(stderr, "A %lf %lf %lf\n", accel[0], accel[1], accel[2]) && 0));

    Update(gyro, R_GYRO, timestamp);
    Update(accel, R_GRAVONLY, timestamp);

    gpsFixer.SetLatestTimestamp(timestamp);
  }
    break;
  case lagrMessageGPS: {
    LagrGPS* gpsP = (LagrGPS*)mdata;
    double gpspos[5], gpsvel[3];
    gpspos[0] = gpsP->latitude;
    gpspos[1] = gpsP->longitude;
    gpspos[2] = gpsP->mslHeight;
    gpspos[3] = gpsP->estPosErrorHor;
    gpspos[4] = gpsP->estPosErrorVert;
    gpsvel[0] = gpsP->velNorth;
    gpsvel[1] = gpsP->velEast;
    gpsvel[2] = gpsP->velUp; 

    assert(!isnan(timestamp));

    Update(gpspos, R_GPSPOS, timestamp);
    Update(gpsvel, R_GPSVEL, timestamp);

    if (exportDisplay) {
      double gpsarray[8] = {
	gpspos[0], gpspos[1], gpspos[2], gpspos[3], gpspos[4],
	gpsvel[0], gpsvel[1], gpsvel[2]
      };
      SendPlotMessage(gpsarray, 8, PLOT_GPS, displayFd);
    }
  }
    break;
  case lagrMessagePosition: {
    LagrPoseState *poseP = (LagrPoseState*)mdata;

    if (exportPose)
      ExportPose(Estimate(), Estimate(), poseP);

    if (exportDisplay) {
      double posearray[7];
      memcpy(posearray, Estimate()->GetPosMean(), sizeof(double)*3);
      Estimate()->GetQuaternion()->GetEulerAngles(posearray + 3);
	    //memcpy(posearray + 3, Estimate()->GetQuaternion()->GetArray(), sizeof(double)*4);
	    //      SendPlotMessage(posearray, 7, PLOT_LOCALPOSE, displayFd);
      SendPlotMessage(posearray, 6, PLOT_POSE, displayFd);
      SendPlotMessage(poseP->globalPose, 6, PLOT_BASELINEPOSE, displayFd);
    }
  }
    break;
  default:
    break;
  }
}

void LagrPoseFilter::InitFilter(LagrPoseState *pose) {
  double roll = pose->globalPose[3];
  double pitch = pose->globalPose[4];
  double yaw = pose->globalPose[5];
  InitialPoseCovar pcovar;
  double delta_rates[3] = { 0.0, 0.0, 0.0 };

  for (int i = 0; i < particleCount; i++) {
    RBParticle *particle = particles[i];

    // set initial rotation to baseline + random rotation 
    Quaternion *newquat = new Quaternion(roll, pitch, yaw);

    for (int del_i = 0; del_i < 3; del_i++) {
      //      delta_rates[del_i] = (drand48() - 0.5) * MIN_INITIAL_ANGLE_UNCERTAINTY;
      // HACK: ignore roll and pitch variance, because it's much more 
      // important to get yaw right
      if (del_i == 2) {
      delta_rates[del_i] = 
	(drand48() - 0.5) * 
	(sqrt(pose->globalPoseCov[del_i + 3] + 0.001) + MIN_INITIAL_ANGLE_UNCERTAINTY);
      } else {
	delta_rates[del_i] = 0.0;
      }
    }

    printf("Covariance: %lf %lf %lf\n", pose->globalPoseCov[3], 
	   pose->globalPoseCov[4], 
	   pose->globalPoseCov[5]
	   );


    newquat->IntegrateRotation(1.0, delta_rates);
    particle->SetNewQuaternion(newquat);

    // copy position, reset covariance
    particle->ZeroPosition();
    memcpy(particle->GetPosMean(), pose->globalPose, sizeof(double) * 3);
    memcpy(particle->GetPosCovar(), pcovar.GetMatrix(), 
	   sizeof(double) * pcovar.Rows() * pcovar.Cols());
  }
}

void LagrPoseFilter::ExportPose(RBParticle *global_estimate, 
				RBParticle *local_estimate,
				LagrPoseState *baseline_pose) {
  RBParticle global_estimate_copy(*global_estimate);
  RBParticle local_estimate_copy(*local_estimate);
  Quaternion *global_quat = global_estimate_copy.GetQuaternion();
  //  Quaternion *local_quat = local_estimate_copy.GetQuaternion();
  double *globalxyz = global_estimate_copy.GetPosMean();
  double *localxyz = local_estimate_copy.GetPosMean();

  // compensate for message lag by projecting estimates forward in time
  global_estimate_copy.TimeUpdateRotation(messageLag);
  local_estimate_copy.TimeUpdateRotation(messageLag);

  for (int i = 0; i < 3; i++) {
    globalxyz[i] += globalxyz[i + 3] * messageLag;
    localxyz[i] += localxyz[i + 3] * messageLag;
  }

  int packetsize = (sizeof(struct lagr_penn_header) + sizeof(LagrPoseState)); 
  char *packet = (char*)malloc(packetsize);

  assert(packet != NULL);

  LagrPoseState *posepacket = 
    (LagrPoseState*)(packet + sizeof(struct lagr_penn_header));

  memcpy(posepacket, baseline_pose, sizeof(LagrPoseState));
  
  memcpy(posepacket->globalPose, globalxyz, sizeof(double)*3);
  global_quat->GetEulerAngles(posepacket->globalPose + 3);
  memcpy(posepacket->globalPoseRate, globalxyz + 3, sizeof(double)*3);
  memcpy(posepacket->globalPoseRate + 3, global_estimate_copy.GetRotationalVelocity(), sizeof(double)*3);

  // CHANGE: SEND BASELINE POSE'S LOCAL POSE ESTIMATE INSTEAD
  // OF DUPLICATING GLOBAL POSE ESTIMATE
  /*
  memcpy(posepacket->localPose, localxyz, sizeof(double)*3);
  local_quat->GetEulerAngles(posepacket->localPose + 3);
  memcpy(posepacket->localPoseRate, localxyz + 3, sizeof(double)*3);
  memcpy(posepacket->localPoseRate + 3, local_estimate_copy.GetRotationalVelocity(), sizeof(double)*3);
  */

  /*
  fprintf(stderr, "XYZ %lf %lf %lf\n", 
	  localxyz[0], localxyz[1], localxyz[2]);
  */

  struct lagr_penn_header *header = (struct lagr_penn_header *)packet;
  header->magicNumber = 0x50454e30;
  header->messageType = lagrPennMessagePose;
  header->timeStamp = posepacket->timeStamp;

  for (unsigned int i = 0; i < destSlots.size(); i++) 
    lagrCommSend((LagrProcessType)destSlots[i],
		 (unsigned char *)packet, 
		 packetsize);

  if (forwardPose) {
    int retval = send(forwardFd, packet, packetsize, MSG_DONTWAIT);
    if (retval == -1) 
      perror("forwardPose");
  }

  free(packet);
}

void LagrPoseFilter::SetForwardHost(const char *host, int port) {
  forwardFd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

  int trueval = 1;

  if (setsockopt(forwardFd, SOL_SOCKET, SO_BROADCAST, &trueval, sizeof(int)) == -1) {
    perror("setsockopt");
    exit(1);
  }

  assert(displayFd != -1);

  struct sockaddr_in destaddr;
  memset(&destaddr, 0, sizeof(struct sockaddr_in));
  destaddr.sin_family = AF_INET;
  destaddr.sin_port = htons(port);
  inet_aton(host, &(destaddr.sin_addr));

  int connectresult = 
    connect(forwardFd, (struct sockaddr*)&destaddr,sizeof(struct sockaddr_in));

  if (connectresult == -1) {
    perror("connect");
    exit(1);
  }

  assert(connectresult != -1);

  forwardPose = true;
}

void LagrPoseFilter::SetDisplayHost(const char *host, int port) {
  displayFd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

  int trueval = 1;

  if (setsockopt(displayFd, SOL_SOCKET, SO_BROADCAST, &trueval, sizeof(int)) == -1) {
    perror("setsockopt");
    exit(1);
  }

  assert(displayFd != -1);

  struct sockaddr_in destaddr;
  memset(&destaddr, 0, sizeof(struct sockaddr_in));
  destaddr.sin_family = AF_INET;
  destaddr.sin_port = htons(port);
  inet_aton(host, &(destaddr.sin_addr));

  int connectresult = 
    connect(displayFd, (struct sockaddr*)&destaddr,sizeof(struct sockaddr_in));

  if (connectresult == -1) {
    perror("connect");
    exit(1);
  }

  assert(connectresult != -1);

  exportDisplay = true;
}

void LagrPoseFilter::SendPlotMessage(double *data, int count, plot_message_t type, int sockfd) {
  PlotMessage pm(type, count * sizeof(double), data);
  char *textmessage = pm.AsString();

  int retval = send(sockfd, textmessage, strlen(textmessage)+1, MSG_DONTWAIT);

  if (retval == -1) 
    perror("send_plot_message");
}

