/*
  lagr/LagrPoseFilter.h

  (C) 2005 Paul Vernaza, <vernaza@stwing.upenn.edu>

  Subclass of PoseRBFilter that contains LAGR-specific things.
*/

#ifndef __LAGR_RBPOSE_H__
#define __LAGR_RBPOSE_H__

#include <vector>
using namespace std;

#include <lagrInterface.h>
#include <string.h>

#include "../plotting/PlotMessage.h"
#include "../PoseRBFilter.h"
#include "../jittercorrector/JitterCorrector.h"
#include "GPSFixer.h"

/* Default delay to use in sorting incoming messages */
#define DEFAULT_MESSAGE_SORTER_LAG 0.08

/* Physical parameters used to calculate velocity from odometry */
#define FRONT_AXLE_LEN 0.6
#define WHEEL_RADIUS 0.174
#define ENCODER_RADIANS_PER_TICK ((2.0*M_PI) / 4096.0)
#define MAX_ENCODER_JUMP 4096

/* It is assumed that the robot will remain stationary for this 
 long after starting in order to estimate the gyro bias. */
#define BIAS_ESTIMATION_TIMELIMIT (5.0)

/* If the magnitude of the estimated bias is greater than this, 
 it will be thrown out */
#define BIAS_MAGNITUDE_LIMIT 0.2

/* The minimum initial uncertainty in angle */
#define MIN_INITIAL_ANGLE_UNCERTAINTY 0.1

#define MAX_LAGR_MESSAGE_SIZE 1024

class LagrMessageWrapper {
public:
  LagrMessageWrapper(LagrMessageType typea, unsigned char *dataa, 
		     size_t size) {
    type = typea;
    assert(size <= MAX_LAGR_MESSAGE_SIZE);
    memcpy(data, dataa, size);
  }

  LagrMessageType type;
  unsigned char data[MAX_LAGR_MESSAGE_SIZE];
};

typedef enum { S_FirstUpdate, S_PoseUninitialized, S_GrabBaselinePose, S_Run } pose_state_t;

class LagrPoseFilter : public PoseRBFilter {
 public:
  LagrPoseFilter(int particles, double message_lag = DEFAULT_MESSAGE_SORTER_LAG);

  /* Queues up a LAGR binary-format message to be sorted and applied later */
  void AddLagrMessage(LagrMessageType type, unsigned char *data);

  /* Randomizes initial rotations */
  void Randomize();

  /* Add this slot to the list of processes that will receive pose messages */
  void AddDestSlot(int slot) { 
    exportPose = true; 
    destSlots.push_back(slot);
  }

  /* Sets address of host to which to send display messages via UDP */
  void SetDisplayHost(const char *host, int port);

  void SetForwardHost(const char *host, int port);

 private:
  void ProcessMessage(double timestamp, LagrMessageWrapper *mwrap);
  void InitFilter(LagrPoseState *pose);

  /* Exports pose to processes specified using AddDestSlot, using LAGR API.
   */
  void ExportPose(RBParticle *global_estimate, 
		  RBParticle *local_estimate,
		  LagrPoseState *baseline_pose);

  /* Sends a UDP message to the plotting program */
  void SendPlotMessage(double *data, int count, plot_message_t type, int sockfd);

  JitterCorrector<LagrMessageWrapper*> msorter;
  double gyroBias[3];
  pose_state_t poseState;
  double startTime;
  double estimatedBias[3];
  double sumGyroReadings[3];
  int biasReadingCount;
  GPSFixer gpsFixer;

  bool exportPose;
  vector<int> destSlots;

  bool exportDisplay;
  int displayFd;

  bool forwardPose;
  int forwardFd;

  double messageLag;
};

#endif
