/*
  lagr/OfflinePoseFilter.cpp

  (C) 2005 Paul Vernaza, <vernaza@stwing.upenn.edu>
*/

#include "OfflinePoseFilter.h"

#include <gsl/gsl_cblas.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/types.h>

#include "lagr_penn.h"
#include "../UTM/LatLong-UTMconversion.h"

OfflinePoseFilter::OfflinePoseFilter(int particlecount, double lag) :
  PoseRBFilter(particlecount),
  msorter_log(lag)
{
  fprintf(stderr, "LAG = %lf\n", lag);

  //  Randomize();

  useGpsp = true;
  useGpsv = true;
  useOd = true;
  useAccel = true;
  useGyro = true;
  useMag = false;
}

void OfflinePoseFilter::Randomize() {
  // randomize initial rotations
  double rotvel[3];
  rotvel[0] = 0.0; rotvel[1] = 0.0; rotvel[2] = 0.0;

  /*
  // flip particles over
  for (int i = 0; i < particleCount; i++) {
    rotvel[0] = 1.0 * M_PI;
    particles[i]->GetQuaternion()->IntegrateRotation(1.0, rotvel);
  }
  */

  rotvel[0] = 0.0; rotvel[1] = 0.0; rotvel[2] = 0.0;

  for (int i = 0; i < particleCount; i++) {
    /*
    rotvel[0] = drand48() * M_PI * 2;
    rotvel[1] = drand48() * M_PI * 2;
    rotvel[2] = drand48() * M_PI * 2;
    */
    rotvel[2] = ((double)i / particleCount) * 2.0 * M_PI;
    particles[i]->GetQuaternion()->IntegrateRotation(1.0, rotvel);
  }
}

void OfflinePoseFilter::SetMeasurementUsage(measurement_t type, bool val) {
  switch (type) {
  case R_GPSPOS:
    useGpsp = val;
    break;
  case R_GPSVEL:
    useGpsv = val;
    break;
  case R_OD:
    useOd = val;
    break;
  case R_GRAVONLY:
    useAccel = val;
    break;
  case R_GYRO:
    useGyro = val;
    break;
  default:
    assert(0);
    break;
  }
}

void OfflinePoseFilter::AddLogMessage(double tstamp, 
				      char type, double *data, int count) {
  LagrLogDatumWrapper *wrapper;
  double new_gps_packet[8];

  switch (type) {
  case 'S': {
    double northingeasting[2];
    LLtoUTMrad(data[0], data[1], northingeasting[0], northingeasting[1]);
    data[2] *= -1;
    data[7] *= -1;
    double gpsarray[] = {
      northingeasting[0], northingeasting[1], data[2], 
      data[5], data[6], data[7]
    };
    gpsFixer.ProjectGPSForward(tstamp, gpsFixer.LastTimestamp(), gpsarray);
    double newgpsdata[] = {
      gpsarray[0], gpsarray[1], gpsarray[2], 
      data[3], data[4], 
      gpsarray[3], gpsarray[4], gpsarray[5]
    };
    memcpy(new_gps_packet, newgpsdata, sizeof(double)*8);
    data = new_gps_packet;
  }
    break;
  case 'A':			// Y and Z need flip
    data[1] = -data[1];
    data[2] = -data[2];
    break;
  default: 
    break;
  }

  wrapper = new LagrLogDatumWrapper(type, data, count);
  if (!msorter_log.EnqueueMessage(tstamp, wrapper)) {
    printf("warning: message not delivered\n");
    delete wrapper;
  }

  double storedtstamp;
  while (msorter_log.GetMessage(&wrapper, &storedtstamp)) {
    double *wdata = wrapper->data;
    gpsFixer.SetLatestTimestamp(storedtstamp);

    switch (wrapper->type) {
    case 'A':
      if (useAccel)
	Update(wdata, R_GRAVONLY, storedtstamp);
    case 'M':
      if (useMag)
	Update(wdata, R_MAG, storedtstamp);
      break;
    case 'G':
      if (useGyro)
	Update(wdata, R_GYRO, storedtstamp);
      break;
    case 'S':
    case 'Z':
      if (useGpsp)
	Update(wdata, R_GPSPOS, storedtstamp);
      if (useGpsv) {
	Update(wdata + 5, R_GPSVEL, storedtstamp);
      }
      break;
    case 'V':
      if (useOd)
	Update(wdata, R_OD, storedtstamp);
      gpsFixer.AddEstimate(storedtstamp, Estimate()->GetPosMean());
      break;
    default:
      break;
    }

    delete wrapper;
  }
}

