/*
  lagr/ocs_plot.cpp

  (C) 2005 Paul Vernaza, <vernaza@stwing.upenn.edu>
*/

#include <arpa/inet.h>
#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <math.h>

#include "../UTM/LatLong-UTMconversion.h"
#include "../Quaternion.h"
#include "../plotting/Plotter.h"
#include "../plotting/PlotMessage.h"
#include "../plotting/IMUGraphic.h"
#include "../plotting/GPSPGraphic.h"
#include "../plotting/PathGraphic.h"
#include "../plotting/LocalMapGraphic.h"

#define MAX_UDP_PACKET_SIZE 2048

//#define LISTEN_PORT 19111
//#define BIND_ADDR "192.168.19.255"

Plotter *plotter;

int Socket_Fd;

IMUGraphic *poseGraphic, *baselinePoseGraphic, *localPoseGraphic;
GPSPGraphic *gpsPosGraphic;
PathGraphic *pathGraphic;
LocalMapGraphic *localMapGraphic;
double currentZ;

void process_readings();
// void init_socket();
void message_handler(user_message_t umess, char *messdata);

//typedef enum { G_POSE, G_BASELINE_POSE, G_GPSPOS, G_LOCALPOSE } graphic_ids_t;

int main(int argc, char *argv[]) {
  if (fcntl(fileno(stdin), F_SETFL, O_NONBLOCK) == -1) {
    perror("fcntl");
    exit(1);
  }

  int window_width = 512, window_height = 384;
  bool enable_aa = false;

  if (argc == 3) {
    window_width = atol(argv[1]);
    window_height = atol(argv[2]);
  }

  if (strstr(argv[0], "_accel")) {
    enable_aa = true;
    window_width = 1024;
    window_height = 768;
  }

  plotter = new Plotter(argc, argv, process_readings, 
			window_width, window_height, enable_aa);
  plotter->RegisterMenuMessageHandler(message_handler);

  poseGraphic = new IMUGraphic("gpose", 2.5f, 1.0, 1.0, 0.0, 1.0);
  localPoseGraphic = new IMUGraphic("lpose", 2.5f, 1.0, 0.5, 0.5, 1.0);
  baselinePoseGraphic = 
    new IMUGraphic("bas_pose", 2.5f, 1.0, 0.5, 0.0, 1.0);
  gpsPosGraphic = new GPSPGraphic();
  pathGraphic = new PathGraphic();
  localMapGraphic = new LocalMapGraphic(0.5);

  poseGraphic->SetTrailLength(10000);
  baselinePoseGraphic->SetTrailLength(1000);

  plotter->AddDrawable(poseGraphic);
  plotter->AddDrawable(baselinePoseGraphic);
  plotter->AddDrawable(gpsPosGraphic);
  //  plotter->AddDrawable(localPoseGraphic);
  plotter->AddDrawable(pathGraphic);
  plotter->AddDrawable(localMapGraphic);

  plotter->Resume();
  plotter->SetSingleStepping(false);
  plotter->SetCamFocus(localPoseGraphic->GetPos());
  plotter->SetCamFocus(poseGraphic->GetPos());
  plotter->Start();
}

void message_handler(user_message_t umess, char *messdata) {
  switch (umess) {
  case (UMESS_ARB): {
    if (strcmp("focus global", messdata) == 0) {
      printf("glob\n");
      plotter->SetCamFocus(poseGraphic->GetPos());
    } else if (strcmp("focus baseline", messdata) == 0) {
      plotter->SetCamFocus(baselinePoseGraphic->GetPos());
    } else if (strcmp("focus local", messdata) == 0) {
      plotter->SetCamFocus(localPoseGraphic->GetPos());
    } else if (strcmp("clear", messdata) == 0) {
    } else {
      printf("Unhandled user message '%s'", messdata);
    }
  }
    break;
  default:
    printf("unhandled message, id = %d\n", umess);
    break;
  }
}

void subtract_bias(double *number) {
  number[0] = int(floor(number[0])) % 10000 + number[0] - floor(number[0]);
  number[1] = int(floor(number[1])) % 10000 + number[1] - floor(number[1]);
}

void process_readings() {
  char packet[1024];

  if (fgets(packet, 1024, stdin) == NULL) {
    usleep(5000);
    return;
  }

  PlotMessage plot_message(packet);
  
  double *mdata = (double*)plot_message.Data();

  //  assert(bytesread == sizeof(PlotMessage));

  switch (plot_message.Type()) {
  case PLOT_GPS: {
    subtract_bias(mdata);

    mdata[2] = 0.0;

    double vel_xyz[3];
    vel_xyz[0] = mdata[5];
    vel_xyz[1] = mdata[6];
    vel_xyz[2] = mdata[7];

    gpsPosGraphic->SetPosErr(mdata, mdata + 3);
    poseGraphic->GetGPSVGraphic()->SetVel(vel_xyz);
  }
    break;
  case PLOT_VEL: {
    poseGraphic->GetVelGraphic()->SetVel(mdata);
    //    printf("vee\n");
  }
    break;
  case PLOT_POSE: {
    /*
    struct timeval tv;
    gettimeofday(&tv, NULL);
    double ctime = tv.tv_sec + 1e-6 * tv.tv_usec;
    fprintf(stderr, "C %lf %lf\n", ctime, mdata[5]);
    */

    subtract_bias(mdata);

    currentZ = mdata[2] = 0.0;

      //    currentZ = mdata[2];
    Quaternion quat(mdata[3],mdata[4],mdata[5]);
    poseGraphic->SetPos(mdata);
    poseGraphic->SetQuat(quat.GetArray());
  }
    break;
  case PLOT_BASELINEPOSE: {
    /*
    struct timeval tv;
    gettimeofday(&tv, NULL);
    double ctime = tv.tv_sec + 1e-6 * tv.tv_usec;
    fprintf(stderr, "B %lf %lf\n", ctime, mdata[5]);

    */

    subtract_bias(mdata);


    mdata[2] = 0.0;


    Quaternion quat(mdata[3],mdata[4],mdata[5]);
    baselinePoseGraphic->SetPos(mdata);
    baselinePoseGraphic->SetQuat(quat.GetArray());
  }
    break;
  case PLOT_LOCALPOSE: {
    /*
    subtract_bias(mdata);
    localPoseGraphic->SetPos(mdata);
    localPoseGraphic->SetQuat(mdata + 3);
    */
  }
    break;
  case PLOT_PATH: {
    for (int i = 0; i < plot_message.DataCount() / 2; i++) 
      subtract_bias(mdata + 2*i);
    pathGraphic->SetPath(currentZ, mdata, plot_message.DataCount() / 2);
  }
    break;
  case PLOT_LOCALMAP: {
    subtract_bias(mdata);
    localMapGraphic->SetCell(mdata, currentZ);
  }
    break;
  default:
    break;
  }

  plotter->Update();
}
