/*
  plotting/Plotter.cpp

  (C) 2005 Paul Vernaza, <vernaza@stwing.upenn.edu>
*/

#include <stdio.h>
#include <glut.h>
#include <assert.h>
#include <math.h>
#include <vector>

using namespace std;

#include "Plotter.h"

#define REDRAW_PERIOD 33

#define GRID_SPAN  500.0
#define GRID_LINE_SPAN_COUNT  500

int Plotter::magicNum = 0;
Plotter *Plotter::currentObject = NULL;

static void redraw(int foo) { glutPostRedisplay(); }

Plotter::Plotter(int argc, char *argv[],
		 void (*idle_func)(void),
		 int width, int height,
		 bool enable_aa) :
  camFilter(0.5),
  menu(width, height)  
{
  assert(Plotter::magicNum != 42);
  Plotter::magicNum = 42;

  singleStepping = true;

  drawableCount = 0;

  stepCounter = 0;
  stepSize = 50;

  currentObject = this;

  idleFunc = idle_func;
  winWidth = width; winHeight = height;

  camChaseDist[0] = 0.0;
  camChaseDist[1] = 0.0;
  camChaseDist[2] = -40.0;

  InitGL(argc, argv, enable_aa);

  gridGraphic = new GridGraphic(GRID_SPAN, GRID_LINE_SPAN_COUNT);

  AddDrawable(gridGraphic);
}

void Plotter::Start() {  
  glutMainLoop();
}

void Plotter::Pause() {
  menu.DispMessage("Paused");
  glutIdleFunc(NULL);
}

void Plotter::Resume() {
  stepCounter = 0;
  glutIdleFunc(idleFunc);
  menu.DispMessage(singleStepping ? "Step" : "Resumed");  
}

void Plotter::Update() {
  if (singleStepping && stepCounter++ == stepSize) {
    Pause();
    stepCounter = 0;
  }
}

void Plotter::InitGL(int argc, char *argv[], bool enable_aa) {
  glutInit(&argc, argv);
  glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  //  glutInitDisplayMode (GLUT_SINGLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize (winWidth, winHeight);
  glutInitWindowPosition (100, 100);
  glutCreateWindow (argv[0]);
  
  glutDisplayFunc(Plotter::wDrawScene);
  glutReshapeFunc(Plotter::wResizeScene);
  glutKeyboardFunc(Plotter::wKeyboardHandler);
  glutMouseFunc(Plotter::wMouseHandler);
  
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f);  
  glClearDepth(1.0f); 
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_DEPTH_TEST);

  if (enable_aa) {
    glEnable(GL_LINE_SMOOTH);
    glEnable(GL_POINT_SMOOTH);
    glEnable(GL_POLYGON_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  }

  glDepthFunc(GL_LEQUAL);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
}

void Plotter::ResizeScene(int width, int height) {
  if (height==0) height=1;
  winWidth = width;
  winHeight = height;
  menu.SetWinSize(width, height);
  
  glViewport(0, 0, width, height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0f,(GLfloat)width/(GLfloat)height,0.1f,10e3f);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}

void Plotter::MouseHandler(int button, int state, int x, int y) {
  int modifier = glutGetModifiers();

  if (state == GLUT_UP) return;

  switch (button) {
  case GLUT_LEFT_BUTTON:
    if (modifier == GLUT_ACTIVE_CTRL) {
      /*
      int viewport[4];
      double mvmatrix[16], projmatrix[16];
      GLdouble worldx, worldy, worldz;
      double est_dist_to_focus = 
	sqrt(camChaseDist[0]*camChaseDist[0] + 
	     camChaseDist[1]*camChaseDist[1] + 
	     camChaseDist[2]*camChaseDist[2]);

      glGetIntegerv(GL_VIEWPORT, viewport);
      glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix);
      glGetDoublev(GL_PROJECTION_MATRIX, projmatrix);
      
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      glTranslated(camFocus[0], camFocus[1], camFocus[2]);

      gluUnProject((GLdouble)x, (GLdouble)(viewport[3] - y - 1),
		   est_dist_to_focus, mvmatrix, projmatrix, viewport, 
		   &worldx, &worldy, &worldz);
      printf("W %lf %lf %lf\n", worldx, worldy, worldz);
      */
    } else {
      double factor = modifier == GLUT_ACTIVE_SHIFT ? 1.25 : 0.75;
      for (int i = 0; i < 3; i++) camChaseDist[i] *= factor;
    }
    break;
  }
}

void Plotter::KeyboardHandler(unsigned char key, int xpos, int ypos) {
  user_message_t umess = menu.HandleKeystroke(key, xpos, ypos);
  char *messdata = menu.GetUserMessage();
  int stepsize;

  switch (umess) {
  case (UMESS_NONE):
    break;
  case (UMESS_CAMPOS): {
    double *chasedist = (double*)messdata;
    chasedist[2] *= -1;
    memcpy(camChaseDist, chasedist, sizeof(double)*3);
  }
    break;
  case (UMESS_ARB): 
    if (sscanf(messdata, "stepsize %d\n", &stepsize) == 1) {
      stepSize = stepsize;
    } else {
      menuHandler(umess, messdata);
    }
    break;
  case (UMESS_CENTERCAM):
    camFilter.SetPos(camFocus);	
    break;
  case (UMESS_SINGLESTEP): 
    singleStepping = !singleStepping;
    menu.DispMessage(singleStepping ? 
		     "Single-stepping: true" : 
		     "Single-stepping: false");
    break;
  case (UMESS_PAUSE):
    Pause();
    break;
  case (UMESS_RESUME):
    Resume();
    break;
  case (UMESS_RESET):
    Pause();
    menu.DispMessage("Reset");
    menuHandler(umess, messdata);
    break;
  case (UMESS_EXIT):
    exit(0);
    break;
  default:
    menuHandler(umess, messdata);
    break;
  }
}

void Plotter::DrawTime() {
  char timebuf[255];
  time_t curtime = (int)currentTime;
  struct tm *loctime = localtime(&curtime);
  strftime(timebuf, 255, "%d %B %Y %r", loctime);
  sprintf(timebuf + strlen(timebuf), " %lf", currentTime - curtime);
  glColor3d(1.0, 1.0, 1.0);
  menu.DrawString(timebuf, 0, int(0.9 * winHeight));
}

void Plotter::DrawScene() {
  glMatrixMode(GL_MODELVIEW);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);  
  glLoadIdentity();

  DrawTime();

  menu.Draw();

  double *filtered_focus = camFilter.Update(camFocus);

  gluLookAt(camChaseDist[0] + filtered_focus[0], 
	    camChaseDist[1] + filtered_focus[1] + 0.0001, 
	    camChaseDist[2] + filtered_focus[2],
	    filtered_focus[0], filtered_focus[1], filtered_focus[2],
	    1.0, 0.0, 0.0);

  double gridCenterX = 
    (GRID_SPAN / 2.0) * floor(camFocus[0] / (GRID_SPAN / 2.0));
  double gridCenterY = 
    (GRID_SPAN / 2.0) * floor(camFocus[1] / (GRID_SPAN / 2.0));
  double gridCenterZ = 10.0 * floor(camFocus[2] / 10.0);

  gridGraphic->SetPos(gridCenterX, gridCenterY, gridCenterZ);

  for (int i = 0; i < drawableCount; i++) {
    drawables[i]->Draw();
  }

  GLenum errcode = glGetError();
  if (errcode != GL_NO_ERROR) {
    printf("GL error code %d\n", (int)errcode);
  }

  //glFlush();
  glutSwapBuffers();
  glutTimerFunc(REDRAW_PERIOD, redraw, 0);
}


